# (c) 2015-2022 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
from os import path, makedirs
import numpy as np
from htmd.adaptive.adaptive import AdaptiveBase
from protocolinterface import val
from htmd.projections.tica import TICA
from htmd.projections.metric import Metric
from htmd.adaptive.util import updatingMean
from htmd.clustering.regular import RegCluster
from sklearn.cluster import MiniBatchKMeans
import unittest
import logging

logger = logging.getLogger(__name__)


class AdaptiveBandit(AdaptiveBase):
    """

    Parameters
    ----------
    app : :class:`SimQueue <jobqueues.simqueue.SimQueue>` object, default=None
        A SimQueue class object used to retrieve and submit simulations
    project : str, default='adaptive'
        The name of the project
    nmin : int, default=0
        Minimum number of running simulations
    nmax : int, default=1
        Maximum number of running simulations
    nepochs : int, default=1000
        Stop adaptive once we have reached this number of epochs
    nframes : int, default=0
        Stop adaptive once we have simulated this number of aggregate simulation frames.
    inputpath : str, default='input'
        The directory used to store input folders
    generatorspath : str, default='generators'
        The directory containing the generators
    dryrun : boolean, default=False
        A dry run means that the adaptive will retrieve and generate a new epoch but not submit the simulations
    updateperiod : float, default=0
        When set to a value other than 0, the adaptive will run synchronously every `updateperiod` seconds
    coorname : str, default='input.coor'
        Name of the file containing the starting coordinates for the new simulations
    lock : bool, default=False
        Lock the folder while adaptive is ongoing
    datapath : str, default='data'
        The directory in which the completed simulations are stored
    filter : bool, default=True
        Enable or disable filtering of trajectories.
    filtersel : str, default='not water'
        Filtering atom selection
    filteredpath : str, default='filtered'
        The directory in which the filtered simulations will be stored
    projection : :class:`Projection <moleculekit.projections.projection.Projection>` object, default=None
        A Projection class object or a list of objects which will be used to project the simulation data before
        constructing a Markov model
    goalfunction : function, default=None
        This function will be used to convert the goal-projected simulation data to a ranking whichcan be used for the
        directed component of FAST.
    reward_method : str, default='max'
        The reward method
    statetype : ('cluster', 'micro', 'macro'), str, default='micro'
        State type (cluster, micro, macro) to use for reward calculations.
    skip : int, default=1
        Allows skipping of simulation frames to reduce data. i.e. skip=3 will only keep every third frame
    lag : int, default=1
        The lagtime used to create the Markov model. Units are in frames.
    exploration : float, default=0.5
        Exploration is the coefficient used in UCB algorithm to weight the exploration value
    temperature : int, default=300
        Temperature used to compute the free energy
    ticalag : int, default=20
        Lagtime to use for TICA in frames. When using `skip` remember to change this accordinly.
    ticadim : int, default=3
        Number of TICA dimensions to use. When set to 0 it disables TICA
    clustmethod : :class:`ClusterMixin <sklearn.base.ClusterMixin>` class, default=<class 'sklearn.cluster._kmeans.MiniBatchKMeans'>
        Clustering algorithm used to cluster the contacts or distances
    macronum : int, default=8
        The number of macrostates to produce
    save : bool, default=False
        Save the model generated
    save_qval : bool, default=False
        Save the Q(a) and N values for every epoch
    actionspace : ('metric', 'goal', 'tica', 'ticapcca'), str, default='tica'
        The projected space which will be clustered and used to define our action space
    recluster : bool, default=False
        If to recluster the projected space used in MSM estimation with a different clustering.
    reclusterMethod : , default=<class 'htmd.clustering.regular.RegCluster'>
        Clustering method for reclustering.
    goal_init : float, default=0.3
        The proportional ratio of goal initialization compared to max frames set by nframes
    """

    def __init__(self):
        from sklearn.base import ClusterMixin
        from moleculekit.projections.projection import Projection

        super().__init__()
        self._arg(
            "datapath",
            "str",
            "The directory in which the completed simulations are stored",
            "data",
            val.String(),
        )
        self._arg(
            "filter",
            "bool",
            "Enable or disable filtering of trajectories.",
            True,
            val.Boolean(),
        )
        self._arg(
            "filtersel", "str", "Filtering atom selection", "not water", val.String()
        )
        self._arg(
            "filteredpath",
            "str",
            "The directory in which the filtered simulations will be stored",
            "filtered",
            val.String(),
        )
        self._arg(
            "projection",
            ":class:`Projection <moleculekit.projections.projection.Projection>` object",
            "A Projection class object or a list of objects which will be used to project the simulation "
            "data before constructing a Markov model",
            None,
            val.Object(Projection),
            nargs="+",
        )
        self._arg(
            "goalfunction",
            "function",
            "This function will be used to convert the goal-projected simulation data to a ranking which"
            "can be used for the directed component of FAST.",
            None,
            val.Function(),
            nargs="any",
        )
        self._arg("reward_method", "str", "The reward method", "mean", val.String())
        self._arg(
            "skip",
            "int",
            "Allows skipping of simulation frames to reduce data. i.e. skip=3 will only keep every third frame",
            1,
            val.Number(int, "POS"),
        )
        self._arg(
            "lag",
            "int",
            "The lagtime used to create the Markov model. Units are in frames.",
            1,
            val.Number(int, "POS"),
        )
        self._arg(
            "exploration",
            "float",
            "Exploration is the coefficient used in UCB algorithm to weight the exploration value",
            0.01,
            val.Number(float, "OPOS"),
        )
        self._arg(
            "temperature",
            "int",
            "Temperature used to compute the free energy",
            300,
            val.Number(int, "POS"),
        )
        self._arg(
            "ticalag",
            "int",
            "Lagtime to use for TICA in frames. When using `skip` remember to change this accordinly.",
            20,
            val.Number(int, "0POS"),
        )
        self._arg(
            "ticadim",
            "int",
            "Number of TICA dimensions to use. When set to 0 it disables TICA",
            3,
            val.Number(int, "0POS"),
        )
        self._arg(
            "clustmethod",
            ":class:`ClusterMixin <sklearn.base.ClusterMixin>` class",
            "Clustering algorithm used to cluster the contacts or distances",
            MiniBatchKMeans,
            val.Class(ClusterMixin),
        )
        self._arg(
            "macronum",
            "int",
            "The number of macrostates to produce",
            8,
            val.Number(int, "POS"),
        )
        self._arg("save", "bool", "Save the model generated", False, val.Boolean())
        self._arg(
            "save_qval",
            "bool",
            "Save the Q(a) and N values for every epoch",
            False,
            val.Boolean(),
        )
        self._arg(
            "actionspace",
            "str",
            "The projected space used to define actions",
            "tica",
            val.String(),
            valid_values=("metric", "goal", "tica", "ticapcca"),
        )
        self._arg(
            "recluster",
            "bool",
            "If to recluster the projected space used in MSM estimation with a different clustering method.",
            False,
            val.Boolean(),
        )
        self._arg(
            "reclusterMethod",
            "",
            "Clustering method used for reclustering.",
            RegCluster,
        )
        self._arg(
            "goal_init",
            "float",
            "The proportional ratio of goal initialization compared to max frames set by nframes",
            0.3,
            val.Number(float, "POS"),
        )
        self._arg(
            "savegoal",
            "str",
            "Save the goal values to the specified file",
            None,
            val.String(),
        )
        self._arg(
            "statetype",
            "str",
            "What states (cluster, micro, macro) to use for calculations.",
            "micro",
            val.String(),
            valid_values=("micro", "cluster", "macro"),
        )

    def _algorithm(self):
        from htmd.kinetics import Kinetics

        sims = self._getSimlist()
        metr = Metric(sims, skip=self.skip)
        metr.set(self.projection)

        data = metr.project()
        data.dropTraj()  # Drop before TICA to avoid broken trajectories

        if self.goalfunction is not None:
            if self.nframes <= 0:
                raise RuntimeError(
                    "When using a goal function, nframes needs to be set to a value greater than 0"
                )
            goaldata = self._getGoalData(data.simlist)
            if len(data.simlist) != len(goaldata.simlist):
                raise RuntimeError(
                    "The goal function was not able to project all trajectories that the MSM projection could. Check for possible errors in the goal function."
                )
            goaldataconcat = np.concatenate(goaldata.dat)
            if self.save:
                makedirs("saveddata", exist_ok=True)
                goaldata.save(
                    path.join("saveddata", f"e{self._getEpoch()}_goaldata.dat")
                )

        # tica = TICA(metr, int(max(2, np.ceil(self.ticalag))))  # gianni: without project it was tooooo slow
        if self.ticadim > 0:
            ticalag = int(
                np.ceil(max(2, min(np.min(data.trajLengths) / 2, self.ticalag)))
            )  # 1 < ticalag < (trajLen / 2)
            tica = TICA(data, ticalag)
            datatica = tica.project(self.ticadim)
            if not self._checkNFrames(datatica):
                return False
            self._createMSM(datatica)
        else:
            if not self._checkNFrames(data):
                return False
            self._createMSM(data)

        confstatdist = self.conformationStationaryDistribution(self._model)
        if self.actionspace == "metric":
            if not data.K:
                data.cluster(
                    self.clustmethod(n_clusters=self._numClusters(data.numFrames))
                )
            data_q = data.copy()
        elif self.actionspace == "goal":
            data_q = goaldata.copy()
        elif self.actionspace == "tica":
            data_q = datatica.copy()
        elif self.actionspace == "ticapcca":
            data_q = datatica.copy()
            for traj in data_q.trajectories:
                traj.cluster = self._model.macro_ofcluster[traj.cluster]
            data_q.K = self._model.macronum

        if self.recluster:
            print(f"Reclustering with {self.reclusterMethod}")
            data_q.cluster(
                self.reclusterMethod(n_clusters=self._numClusters(data.numFrames))
            )

        numstates = data_q.K
        print(f"Numstates: {numstates}")
        currepoch = self._getEpoch()
        q_values = np.zeros(numstates, dtype=np.float32)
        n_values = np.zeros(numstates, dtype=np.int32)

        if self.goalfunction is not None:
            # For every cluster in data_q, get the max score and initialize
            qstconcat = np.concatenate(data_q.St)
            statemaxes = np.zeros(numstates)
            np.maximum.at(statemaxes, qstconcat, np.squeeze(goaldataconcat))
            goalscale = self._featScale(statemaxes)
            goalenergies = -Kinetics._kB * self.temperature * np.log(1 - goalscale)
            q_values = goalenergies
            n_values += int(
                (self.nframes / self._numClusters(self.nframes)) * self.goal_init
            )  # Needs nframes to be set properly!!!!!!!!

        rewardtraj = np.arange(
            data_q.numTrajectories
        )  # Recalculate reward for all states
        rewards = self.getRewards(
            rewardtraj, data_q, confstatdist, numstates, self.reward_method
        )
        for i in range(numstates):
            if len(rewards[i]) == 0:
                continue
            q_values[i] = updatingMean(q_values[i], n_values[i], rewards[i])
        n_values += np.array([len(x) for x in rewards])

        if self.save_qval:
            makedirs("saveddata", exist_ok=True)
            np.save(path.join("saveddata", f"e{currepoch}_qval.npy"), q_values)
            np.save(path.join("saveddata", f"e{currepoch}_nval.npy"), n_values)

        total_steps = np.sum(n_values)
        ucb_values = np.array(
            [
                self.count_ucb(
                    q_values[clust], self.exploration, total_steps + 1, n_values[clust]
                )
                for clust in range(numstates)
            ]
        )

        if self.save_qval:
            makedirs("saveddata", exist_ok=True)
            np.save(path.join("saveddata", f"e{currepoch}_ucbvals.npy"), ucb_values)

        N = self.nmax - self._running

        action = np.argsort(-ucb_values)[:N]
        action_sel = np.zeros(numstates, dtype=int)
        action_sel[action] += 1

        while np.sum(action_sel) < N:  # When K is lower than N repeat some actions
            for a in action:
                action_sel[a] += 1
                if np.sum(action_sel) == N:
                    break

        if self.save_qval:
            np.save(path.join("saveddata", f"e{currepoch}_actions.npy"), action_sel)
        relFrames = self._getSpawnFrames_UCB(action_sel, data_q)
        self._writeInputs(data.rel2sim(np.concatenate(relFrames)))
        return True

    def _getSimlist(self):
        from glob import glob
        from htmd.simlist import simlist, simfilter

        logger.info("Postprocessing new data")

        sims = simlist(
            glob(path.join(self.datapath, "*", "")),
            glob(path.join(self.inputpath, "*", "")),
            glob(path.join(self.inputpath, "*", "")),
        )

        if self.filter:
            sims = simfilter(sims, self.filteredpath, filtersel=self.filtersel)
        return sims

    def count_ucb(self, q_value, exploration, step, n_value):
        return q_value + (exploration * np.sqrt((np.log(step) / (n_value + 1))))

    def count_pucb(self, q_value, exploration, predictor, step, n_value):
        return q_value + (
            exploration * predictor * np.sqrt((np.log(step) / (n_value + 1)))
        )

    def getRewards(self, trajidx, data_q, confstatdist, numstates, rewardmethod):
        from htmd.kinetics import Kinetics
        import pandas as pd

        rewards = [[] for _ in range(numstates)]
        for simidx in trajidx:
            # Get the eq distribution of each of the states the sim passed through
            states = data_q.St[simidx]
            statprob = confstatdist[simidx]
            connected = (states != -1) & (statprob != 0)
            if not np.any(connected):
                continue
            states = states[connected]
            statprob = statprob[connected]
            energies = -Kinetics._kB * self.temperature * np.log(1 - statprob)

            ww = len(energies)

            if rewardmethod == "mean":
                windowedreward = (
                    pd.Series(energies[::-1])
                    .rolling(ww, min_periods=1)
                    .mean()
                    .values[::-1]
                )
            elif rewardmethod == "max":
                windowedreward = (
                    pd.Series(energies[::-1])
                    .rolling(ww, min_periods=1)
                    .max()
                    .values[::-1]
                )
            else:
                raise RuntimeError(f"Reward method {rewardmethod} not available")

            for st, re in zip(states, windowedreward):
                rewards[st].append(re)

        return rewards

    def conformationStationaryDistribution(self, model):
        statdist = np.zeros(model.data.numFrames)  # zero for disconnected set
        dataconcatSt = np.concatenate(model.data.St)
        if self.statetype == "macro":
            for i in range(model.macronum):
                macroframes = np.where(model.macro_ofcluster[dataconcatSt] == i)[0]
                statdist[macroframes] = model.eqDistribution(plot=False)[i]
        elif self.statetype == "micro":
            for i in range(model.micronum):
                microframes = np.where(model.micro_ofcluster[dataconcatSt] == i)[0]
                statdist[microframes] = model.msm.stationary_distribution[i]
        return model.data.deconcatenate(statdist)

    def _checkNFrames(self, data):
        if self.nframes != 0 and data.numFrames >= self.nframes:
            logger.info("Reached maximum number of frames. Stopping adaptive.")
            return False
        return True

    def _getGoalData(self, sims):
        from htmd.projections.metric import Metric
        from htmd.simlist import _simName
        import pickle
        import os

        logger.debug("Starting projection of directed component")
        metr = Metric(sims, skip=self.skip)
        metr.set(self.goalfunction)
        data = metr.project()
        logger.debug("Finished calculating directed component")

        if self.savegoal is not None:
            dirname = os.path.dirname(self.savegoal)
            os.makedirs(dirname, exist_ok=True)
            savedata = {}
            for traj in data.trajectories:
                trajname = _simName(traj.sim.trajectory[0])
                savedata[trajname] = traj.projection
            with open(self.savegoal, "wb") as f:
                pickle.dump(savedata, f)

        return data

    def _featScale(self, feat):
        denom = np.max(feat) - np.min(feat)
        if denom == 0:  # Handling the trivial case where all states have equal features
            res = np.zeros(len(feat))
            res[:] = 1 / len(feat)
            return res
        return (feat - np.min(feat)) / denom

    def _createMSM(self, data):
        from htmd.model import Model

        kmeanserror = True
        while kmeanserror:
            try:
                data.cluster(
                    self.clustmethod(n_clusters=self._numClusters(data.numFrames))
                )
            except IndexError:
                continue
            kmeanserror = False

        self._model = Model(data)
        self._model.markovModel(self.lag, self._numMacrostates(data))
        if self.save:
            makedirs("saveddata", exist_ok=True)
            self._model.save(
                path.join("saveddata", f"e{self._getEpoch()}_adapt_model.dat")
            )

    def _getSpawnFrames_UCB(self, reward, data):
        stateIdx = np.where(reward > 0)[0]
        _, relFrames = data.sampleClusters(
            stateIdx, reward[stateIdx], replacement=True, allframes=False
        )
        logger.debug(f"relFrames {relFrames}")
        return relFrames

    def _numClusters(self, numFrames):
        """Heuristic that calculates number of clusters from number of frames"""
        K = int(
            max(np.round(0.6 * np.log10(numFrames / 1000) * 1000 + 50), 100)
        )  # heuristic
        if K > numFrames / 3:  # Ugly patch for low-data regimes ...
            K = int(numFrames / 3)
        return K

    def _numMacrostates(self, data):
        """Heuristic for calculating the number of macrostates for the Markov model"""
        macronum = self.macronum
        if data.K < macronum:
            macronum = int(np.ceil(data.K / 2))
            logger.warning(
                f"Using less macrostates than requested due to lack of microstates. macronum = {macronum}"
            )

        # Calculating how many timescales are above the lag time to limit number of macrostates
        from deeptime.util.validation import implied_timescales
        from htmd.model import Model

        statelist = [traj.cluster for traj in data.trajectories]
        model = Model._get_model(statelist, self.lag)
        its_data = implied_timescales(model, n_its=macronum)
        timesc = its_data._its
        macronum = min(self.macronum, max(np.sum(timesc > self.lag), 2))
        return macronum


class _TestAdaptiveBandit(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        from htmd.util import tempname
        from htmd.home import home
        import shutil
        import os

        tmpdir = tempname()
        shutil.copytree(home(dataDir="adaptive"), tmpdir)
        os.chdir(tmpdir)
        print(f"Running AdaptiveBandit test in {tmpdir}")

    # def test_adaptive(self):
    #     from sklearn.cluster import MiniBatchKMeans
    #     from jobqueues.localqueue import LocalCPUQueue
    #     from moleculekit.projections.metricdistance import MetricDistance

    #     import numpy as np
    #     import random

    #     np.random.seed(0)  # Needed for the clustering to always give same results
    #     random.seed(0)

    #     md = AdaptiveBandit()
    #     md.app = LocalCPUQueue()
    #     md.generatorspath = "generators"
    #     md.inputpath = "input"
    #     md.datapath = "data"
    #     md.coorname = "input.coor"
    #     md.filter = True
    #     md.filtersel = "all"

    #     md.clustmethod = MiniBatchKMeans
    #     md.projection = MetricDistance(
    #         "protein resid 173 and name CA",
    #         "resname BEN and name C1 C2 C3 C7",
    #         periodic="selections",
    #     )
    #     md.ticadim = 2
    #     md.nmin = 1
    #     md.nmax = 2
    #     md.nepochs = 9999
    #     md.nframes = 1000000

    #     md.reward_method = "mean"
    #     md.exploration = 0.01
    #     md.actionspace = "tica"
    #     md.actionpool = 0
    #     md.recluster = False

    #     md.save = True
    #     md.dryrun = True
    #     md.run()


if __name__ == "__main__":
    unittest.main(verbosity=2)



# (c) 2015-2022 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import htmd
import os
import sys
import inspect
import platform


def home(dataDir=None, libDir=False, shareDir=False):
    """Return the pathname of the HTMD root directory (or a data subdirectory).

    Parameters
    ----------
    dataDir : str
        If not None, return the path to a specific data directory
    libDir : bool
        If True, return path to the lib directory

    Returns
    -------
    dir : str
        The directory

    Example
    -------
    >>> from htmd.home import home
    >>> home()                                 # doctest: +ELLIPSIS
    '.../htmd'
    >>> home(dataDir="dhfr")                   # doctest: +ELLIPSIS
    '.../data/dhfr'
    >>> os.path.join(home(dataDir="dhfr"),"dhfr.pdb")  # doctest: +ELLIPSIS
    '.../data/dhfr/dhfr.pdb'
    """

    homeDir = os.path.dirname(inspect.getfile(htmd))
    try:
        if sys._MEIPASS:
            homeDir = sys._MEIPASS
    except Exception:
        pass

    if dataDir:
        return os.path.join(homeDir, "data", dataDir)
    elif libDir:
        libdir = os.path.join(homeDir, "lib", platform.system())
        if not os.path.exists(libdir):
            raise FileNotFoundError("Could not find libs.")
        return libdir
    elif shareDir:
        sharedir = os.path.join(homeDir, "share")
        if not os.path.exists(sharedir):
            raise FileNotFoundError("Could not find HTMD share directory.")
        return sharedir
    else:
        return homeDir


if __name__ == "__main__":
    import doctest

    doctest.testmod()

    h = home()
    print(h)



# (c) 2015-2022 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
from htmd.versionwarnings import _issueWarnings, _disableWarnings
import __main__ as main
import os.path
from htmd.config import _config
import htmd.home
import logging.config
from importlib.metadata import version, PackageNotFoundError


try:
    __version__ = version("htmd")
except PackageNotFoundError:
    try:
        __version__ = version("acellera-htmd")
    except PackageNotFoundError:
        __version__ = "0.0.0"


try:
    logging.config.fileConfig(
        os.path.join(htmd.home.home(), "logging.ini"), disable_existing_loggers=False
    )
except Exception:
    print("HTMD: Logging setup failed")

reg_file = os.path.join(
    os.path.expanduser("~"), ".htmd", ".registered-htmd", "registration"
)
if (not hasattr(main, "__file__")) and (
    (not os.path.isfile(reg_file))
    or os.getenv("LICENCE_ACCEPTED") == "YES"
    or os.getenv("CI")
    or os.getenv("APPTAINER_CONTAINER")
):
    print(
        "\nCopyright by Acellera Ltd. By executing you are accepting the License. In order to register, "
        "run htmd_register on your terminal."
    )
    print("The registration information must be valid so that it might be verified.")

if _config["configfile"]:
    if not os.path.isfile(_config["configfile"]):
        raise FileNotFoundError(
            "HTMD Config file {} does not exist.".format(_config["configfile"])
        )
    elif not _config["configfile"].endswith(".py"):
        raise Warning(
            "HTMD Config file {} may not be a python file.".format(
                _config["configfile"]
            )
        )
    else:
        try:
            with open(_config["configfile"]) as f:
                exec(compile(f.read(), _config["configfile"], "exec"))
        except Exception:
            raise RuntimeError(
                "Failed to execute the HTMD Config file {}.".format(
                    _config["configfile"]
                )
            )
        else:
            print("\nHTMD Config file {} executed.".format(_config["configfile"]))

_issueWarnings()



# (c) 2015-2022 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import numpy as np
from htmd.units import convert as unitconvert
import logging

logger = logging.getLogger(__name__)


class GWPCA(object):
    """Class for calculating the globally-weighted PCA projections of a MetricData  object

    References
    ----------

    * N. Blöchliger, A. Caflisch, and A. Vitalis. Weighted Distance Functions Improve Analysis of High-Dimensional
      Data: Application to Molecular Dynamics Simulations, J. Chem. Theory Comput., 2015, 11 (11), pp 5481–5492. doi:
      `10.1021/acs.jctc.5b00618`_

    .. _10.1021/acs.jctc.5b00618: http://dx.doi.org/10.1021/acs.jctc.5b00618

    Parameters
    ----------
    data : :class:`MetricData <htmd.metricdata.MetricData>` object
        The object whose data we wish to project.

    Example
    -------
    >>> gw = GWPCA(data)
    >>> dataproj = gw.project(5)
    """

    def __init__(self, data, lag, units="frames"):
        lag = unitconvert(units, "frames", lag, data.fstep)
        if lag == 0:
            raise RuntimeError(
                "Lag time conversion resulted in 0 frames. Please use a larger lag-time for TICA."
            )
        self.data = data

        datconcat = np.concatenate(self.data.dat)
        self.weights = self._autocorrelation(datconcat, lag)

    def _autocorrelation(self, data, lag):
        # Calculate the autocorrelation of each feature
        autocorr = []
        for i in range(data.shape[1]):
            autocorr.append(np.correlate(data[0:-lag, i], data[lag:, i]))
        return np.squeeze(autocorr)

    def project(self, ndim=None):
        """Projects the data object given to the constructor onto `ndim` dimensions

        Parameters
        ----------
        ndim : int
            The number of dimensions we want to project the data on.

        Returns
        -------
        dataTica : :class:`MetricData <htmd.metricdata.MetricData>` object
            A new :class:`MetricData <htmd.metricdata.MetricData>` object containing the projected data

        Example
        -------
        >>> gw = GWPCA(data)
        >>> dataproj = gw.project(5)
        """
        from sklearn.decomposition import IncrementalPCA
        from tqdm import tqdm

        pca = IncrementalPCA(n_components=ndim, batch_size=10000)
        for t in tqdm(self.data.trajectories):
            pca.partial_fit(t.projection * self.weights)

        projdata = self.data.copy()
        for i, t in enumerate(tqdm(self.data.trajectories)):
            projdata.trajectories[i].projection = pca.transform(
                t.projection * self.weights
            )

        # projdataconc = pca.fit_transform(self.weighedconcat)
        # projdata.dat = projdata.deconcatenate(projdataconc)
        return projdata



# (c) 2015-2022 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import pickle
import logging

logger = logging.getLogger(__name__)


def htmdsave(fname, varnames=None):
    """Saves the workspace to a file

    Parameters
    ----------
    fname : str
        The file in which to save the workspace
    varnames : list
        A list of variables to save. Default: saves all variables.
    """
    from types import ModuleType, FunctionType
    import __main__ as _main_module
    import inspect

    mydict = dict()
    maindict = _main_module.__dict__.copy()
    bannedvars = ["In", "Out", "exit", "quit", "get_ipython"]

    if varnames is None:
        logger.warning(
            "Save will ignore all variables starting with _ and following variable names {}".format(
                bannedvars
            )
        )
    logger.warning(
        "Save will duplicate data which is stored in various pointers or variables. Take care!"
    )

    if varnames is None:
        for k in maindict:
            if (
                k[0] != "_"
                and k not in bannedvars
                and not isinstance(maindict[k], ModuleType)
                and not isinstance(maindict[k], FunctionType)
                and not inspect.isclass(maindict[k])
            ):
                try:
                    pickle.dumps(maindict[k])  # This probably slows down everything
                    mydict[k] = maindict[k]
                except Exception:
                    logger.debug("Skipping variable: {}".format(k))
    else:
        if not isinstance(varnames, list):
            varnames = [varnames]
        for k in varnames:
            mydict[k] = maindict[k]

    logger.debug("Saving variables: {}".format(list(mydict.keys())))
    f = open(fname, "wb")
    pickle.dump(mydict, f)
    f.close()


def htmdload(fname):
    """Loads a previously saved workspace from a file.

    Parameters
    ----------
    fname : str
        The location of the file containing the saved workspace.
    """
    import __main__ as _main_module

    f = open(fname, "rb")
    mydict = pickle.load(f)
    f.close()

    # Inserting all variables into the main dictionary
    _main_module.__dict__.update(mydict)



# (c) 2015-2022 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import tempfile
import numpy as np
from os import path
from htmd.home import home
import os
import shutil
from subprocess import call
from natsort import natsorted
from moleculekit.molecule import Molecule
from networkx.algorithms import isomorphism
from glob import glob
import platform
import logging

logger = logging.getLogger(__name__)


def dock(
    protein,
    ligand,
    center=None,
    extent=None,
    numposes=20,
    babelexe="obabel",
    vinaexe=None,
):
    """Molecular docking, using Vina

    If centre and extent are not provided, docking will be performed over the whole protein

    Parameters
    ----------
    protein : :class:`Molecule <moleculekit.molecule.Molecule>` object
        Molecule object representing the receptor
    ligand : :class:`Molecule <moleculekit.molecule.Molecule>` object
        Molecule object representing the ligand to dock
    center : list
        3-vec centre of of the search bounding box (optional)
    extent : list
        3-vec linear extent of the search bounding box (optional)
    numposes : int
        Number of poses to return. Vina cannot return more than 20 poses.
    babelexe : str
        Path to babel executable.
    vinaexe : str
        Path to AutoDock Vina executable.

    Returns
    -------
    poses
        Molecule object representing the N<10 best poses
    scores
        3x num_poses matrix containing kcal, rmsd lb, rmsd ub

    Examples
    --------
    >>> poses, scoring = dock(protein, ligand)
    >>> poses, scoring = dock(protein, ligand, center=[ 10., 5., 12. ], extent=[ 15., 15., 15. ] )

    """
    if np.size(protein.coords, 2) != 1 or np.size(ligand.coords, 2) != 1:
        raise NameError("Protein and ligand Molecules should be single frames")

    buffer = 10.0  # Angstrom buffer around protein for whole-protein docking
    c_min = np.min(protein.coords, 0).reshape((1, 3))[0]
    c_max = np.max(protein.coords, 0).reshape((1, 3))[0]

    if center is None:
        center = (buffer + (c_max + c_min)) / 2
    if extent is None:
        extent = (c_max - c_min) + buffer

    # babel -i pdb protein.pdb  -o pdbqt protein.pdbqt -xr
    # babel -i pdb ligand.pdb   -o pdbqt ligand.pdbqt -xhn
    # vina --ligand ligand.pdbqt --receptor protein.pdbqt --center_x 0. --center_y 0. --center_z 0. --size_x 60. --size_y 60. --size_z 60 --exhaustiveness 10
    # babel -m -i pdbqt ligand_out.pdbqt -o pdb out_.pdb -xhn

    with tempfile.TemporaryDirectory() as tmpdir:
        protein_pdb = os.path.join(tmpdir, "protein.pdb")
        ligand_mol2 = os.path.join(tmpdir, "ligand.mol2")

        protein.write(protein_pdb)
        lig2 = ligand.copy()
        # babel does not understand mol2 atomtypes and requires elements instead
        lig2.atomtype = lig2.element
        lig2.write(ligand_mol2)

        # Dirty hack to remove the 'END' line from the PDBs since babel hates it
        with open(protein_pdb, "r") as f:
            lines = f.readlines()
        with open(protein_pdb, "w") as f:
            f.writelines(lines[:-1])
        # End of dirty hack

        try:
            if vinaexe is None:
                suffix = ""
                if platform.system() == "Windows":
                    suffix = ".exe"
                vinaexe = f"{platform.system()}-vina{suffix}"

            vinaexe = shutil.which(vinaexe, mode=os.X_OK)
            if not vinaexe:
                raise NameError(
                    "Could not find vina, or no execute permissions are given"
                )
        except Exception:
            raise NameError("Could not find vina, or no execute permissions are given")
        try:
            babelexe = shutil.which(babelexe, mode=os.X_OK)
            if babelexe is None:
                raise NameError(
                    "Could not find babel, or no execute permissions are given"
                )
        except Exception:
            raise NameError("Could not find babel, or no execute permissions are given")

        protein_pdbqt = os.path.join(tmpdir, "protein.pdbqt")
        call(
            [
                babelexe,
                "-i",
                "pdb",
                protein_pdb,
                "-o",
                "pdbqt",
                "-O",
                protein_pdbqt,
                "-xr",
            ]
        )

        ligand_pdbqt = os.path.join(tmpdir, "ligand.pdbqt")
        if np.all(ligand.charge != 0):
            logger.info("Charges detected in ligand and will be used for docking.")
            call(
                [
                    babelexe,
                    "-i",
                    "mol2",
                    ligand_mol2,
                    "-o",
                    "pdbqt",
                    "-O",
                    ligand_pdbqt,
                    "-xn",
                    "-xh",
                ]
            )
        else:
            logger.info(
                "Charges were not defined for all atoms. Will guess charges anew using gasteiger method."
            )
            call(
                [
                    babelexe,
                    "-i",
                    "mol2",
                    ligand_mol2,
                    "-o",
                    "pdbqt",
                    "-O",
                    ligand_pdbqt,
                    "-xn",
                    "-xh",
                    "--partialcharge",
                    "gasteiger",
                ]
            )

        if not path.isfile(ligand_pdbqt):
            raise NameError("Ligand could not be converted to PDBQT")
        if not path.isfile(protein_pdbqt):
            raise NameError("Protein could not be converted to PDBQT")

        output_pdbqt = os.path.join(tmpdir, "output.pdbqt")
        call(
            [
                vinaexe,
                "--receptor",
                protein_pdbqt,
                "--ligand",
                ligand_pdbqt,
                "--out",
                output_pdbqt,
                "--center_x",
                str(center[0]),
                "--center_y",
                str(center[1]),
                "--center_z",
                str(center[2]),
                "--size_x",
                str(extent[0]),
                "--size_y",
                str(extent[1]),
                "--size_z",
                str(extent[2]),
                "--num_modes",
                str(numposes),
            ]
        )

        call(
            [
                babelexe,
                "-m",
                "-i",
                "pdbqt",
                output_pdbqt,
                "-o",
                "pdb",
                "-O",
                os.path.join(tmpdir, "output_.pdb"),
                "-xhn",
            ]
        )

        outfiles = natsorted(glob(os.path.join(tmpdir, "output_*.pdb")))

        lig_g = ligand.toGraph()
        scoring = []
        poses = []
        for i, ligf in enumerate(outfiles):
            scoring.append(_parseScoring(ligf))
            pose = Molecule(ligf)
            pose.viewname = f"Pose {i}"
            pose.resname[:] = ligand.resname[0]

            gm = isomorphism.GraphMatcher(lig_g, pose.toGraph())
            if not gm.is_isomorphic():
                raise RuntimeError(
                    "Ligand molecules were changed and cannot be matched to original molecule"
                )

            # Rename atoms in reference molecule and copy formal charges
            for key, val in gm.mapping.items():
                pose.name[val] = ligand.name[key]
                pose.formalcharge[val] = ligand.formalcharge[key]
            poses.append(pose)

    return poses, np.array(scoring)


def _parseScoring(outf):
    kcal = rmsdlb = rmsdub = None
    with open(outf, "r") as f:
        for line in f:
            if line.startswith("REMARK VINA RESULT:"):
                pieces = line.split()
                kcal = float(pieces[3])
                rmsdlb = float(pieces[4])
                rmsdub = float(pieces[5])
                break
    if kcal is None or rmsdlb is None or rmsdub is None:
        raise RuntimeError("Could not parse vina output correctly {}".format(outf))
    return kcal, rmsdlb, rmsdub


if __name__ == "__main__":
    protein = Molecule(path.join(home(), "data", "docking", "protein.pdb"))
    ligand = Molecule(path.join(home(), "data", "docking", "ligand.pdb"))
    # poses, scoring = dock(protein, ligand)



# (c) 2015-2022 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
from moleculekit.molecule import Molecule
from glob import glob
import numpy as np
import unittest
import logging

logger = logging.getLogger(__name__)

# Based on: http://onlinelibrary.wiley.com/doi/10.1002/(SICI)1097-0134(199601)24:1%3C92::AID-PROT7%3E3.0.CO;2-Q/epdf
# Structure, energetics, and dynamics of lipid–protein interactions: A molecular dynamics study
# of the gramicidin A channel in a DMPC bilayer


class _Lipid:
    def __init__(
        self,
        resname=None,
        headname=None,
        mol=None,
        xyz=None,
        rot=None,
        neighbours=None,
        rings=None,
        area=None,
    ):
        self.resname = resname
        self.headname = headname
        self.mol = mol
        self.xyz = xyz
        self.rot = rot
        self.neighbours = neighbours
        self.rings = rings
        self.area = area

    def __repr__(self):
        return "<{}.{} object at {} {}>".format(
            self.__class__.__module__,
            self.__class__.__name__,
            hex(id(self)),
            self.__str__(),
        )

    def __str__(self):
        s = ""
        if self.resname is not None:
            s += "resname: {} ".format(self.resname)
        if self.headname is not None:
            s += "headname: {} ".format(self.headname)
        if self.xyz is not None:
            s += "xyz: {} ".format(self.xyz)
        if self.neighbours is not None:
            s += "neigh: {} ".format(len(self.neighbours))
        if self.mol is not None:
            s += "mol: {} ".format(id(self.mol))
        return s[:-1]


def listLipids():
    """Lists all available lipids

    Examples
    --------
    >>> from htmd.membranebuilder.build_membrane import buildMembrane
    >>> build_membrane.listLipids()
    ---- Lipids list: ...

    """

    from htmd.home import home
    import os
    from natsort import natsorted

    membranebuilderhome = os.path.join(
        home(shareDir=True), "membranebuilder", "lipids", ""
    )
    lipids = natsorted(glob(os.path.join(membranebuilderhome, "*", "")))
    print("---- Lipids list: " + membranebuilderhome + " ----")
    for ll in lipids:
        print("- ", os.path.basename(os.path.abspath(ll)))
    print("* Lipid DB file: " + os.path.join(membranebuilderhome, "lipiddb.csv"))


def _createLipids(lipidratio, area, lipiddb, files, leaflet=None):
    lipiddb = lipiddb.to_dict(orient="index")
    lipidnames = list(lipidratio.keys())
    ratiosAPL = np.array(
        [lipidratio[lipn] * lipiddb[lipn]["APL"] for lipn in lipidnames]
    )
    # Calculate the total areas per lipid type
    areaspl = area * (ratiosAPL / ratiosAPL.sum())
    # Calculate the counts from the total areas
    counts = np.round(
        areaspl / np.array([lipiddb[lipn]["APL"] for lipn in lipidnames])
    ).astype(int)

    lipids = []
    for i in range(len(lipidnames)):
        resname = lipidnames[i]
        rings = _detectRings(Molecule(files[resname][0]))
        for k in range(counts[i]):
            if leaflet == "upper":
                xyz = np.array([np.nan, np.nan, lipiddb[resname]["Thickness"] / 2])
            elif leaflet == "lower":
                xyz = np.array([np.nan, np.nan, -lipiddb[resname]["Thickness"] / 2])
            lipids.append(
                _Lipid(
                    resname=resname,
                    headname=lipiddb[resname]["Head"],
                    rings=rings,
                    area=lipiddb[resname]["APL"],
                    xyz=xyz,
                )
            )
    return lipids


def _setPositionsLJSim(width, lipids):
    from htmd.membranebuilder.ljfluid import distributeLipids

    sigmas = np.array([2 * np.sqrt(ll.area / np.pi) for ll in lipids])
    resnames = [ll.resname for ll in lipids]

    cutoff = min(np.min(width) / 2, 3 * np.max(sigmas))
    pos = distributeLipids(width + [2 * cutoff], resnames, sigmas, cutoff=cutoff)
    for i in range(len(lipids)):
        lipids[i].xyz[:2] = pos[i, :2]


def _createMembraneMolecule(lipids):
    from moleculekit.util import rotationMatrix

    allmols = []
    numAtoms = 0
    for i, l in enumerate(lipids):
        mol = l.mol.copy()
        headpos = mol.coords[mol.name == l.headname].flatten()[np.newaxis, :]
        mol.moveBy(-headpos)
        mol.rotateBy(rotationMatrix([0, 0, 1], np.deg2rad(l.rot)))
        mol.moveBy(l.xyz)
        mol.resid[:] = i
        numAtoms += mol.numAtoms
        allmols.append(mol)

    # Merge all the lipids into a single Molecule
    mol = Molecule().empty(numAtoms)
    mol.coords = np.zeros((numAtoms, 3, 1), dtype=np.float32)
    start_idx = 0
    for mm in allmols:
        for prop in ["name", "resname", "resid", "segid", "chain", "element", "coords"]:
            mol.__dict__[prop][start_idx : start_idx + mm.numAtoms] = getattr(mm, prop)
        start_idx += mm.numAtoms

    return mol


def _detectRings(mol):
    import networkx as nx

    bonds = mol._guessBonds()

    G = nx.Graph()
    G.add_edges_from(bonds)
    cycles = nx.cycle_basis(G)
    if len(cycles) == 0:
        return None

    fivesix = [c for c in cycles if len(c) in (5, 6)]
    return fivesix


def wrapping_dist_python(coor1, coor2, box):
    assert (coor1.ndim == 1) or (coor2.ndim == 1)
    dist = coor1 - coor2
    dist = dist - box * np.round(dist / box)
    return np.sqrt(np.sum(dist * dist, 1))


def _findNeighbours(lipids, box):
    xypos = np.vstack([ll.xyz[:2] for ll in lipids])

    for i in range(len(lipids)):
        dist = wrapping_dist_python(xypos[i, :], xypos[i + 1 :, :], box)
        lipids[i].neighbours = i + 1 + np.where(dist < 11)[0]


def _loadMolecules(lipids, files):
    from moleculekit.util import rotationMatrix

    # Create Molecules
    for ll in lipids:
        randidx = np.random.randint(len(files[ll.resname]))
        mol = Molecule(files[ll.resname][randidx])
        mol.filter("not water", _logger=False)
        if ll.xyz[2] < 0:
            mol.rotateBy(
                rotationMatrix([1, 0, 0], np.deg2rad(180))
            )  # Rotate the lower leaflet lipids upside down
        ll.mol = mol
        ll.rot = np.random.random() * 360 - 180  # Random starting rotation


def _locateLipidFiles(folder, lipidnames):
    import os

    files = {}
    for mm in lipidnames:
        files[mm] = glob(os.path.join(folder, mm, "*.pdb"))
        if len(files[mm]) == 0:
            raise RuntimeError(
                f'Could not locate pdb files for lipid "{mm}" in folder {folder}'
            )
    return files


def buildMembrane(
    xysize,
    ratioupper,
    ratiolower,
    waterbuff=20,
    platform="CUDA",
    minimize=0,
    equilibrate=0,
    outdir=None,
    lipidf=None,
    ff=None,
    topo=None,
    param=None,
    build=False,
):
    """Construct a membrane containing arbitrary lipids and ratios of them.

    Parameters
    ----------
    xysize : list
        A list containing the size in x and y dimensions of the membrane in Angstroms
    ratioupper : dict
        A dict with keys the molecule names and the ratio of that molecule for the upper layer
    ratiolower : dict
        Same as ratioupper but for the lower layer
    waterbuff : float
        The z-dimension size of the water box above and below the membrane
    platform : str
        The platform on which to run the minimization ('CUDA' or 'CPU')
    minimize : int
        If not 0 it minimizes the membrane for the given number of steps
    equilibrate : float
        If not 0 it equilibrates the membrane for the given number of nanoseconds
    outdir : str
        A folder in which to store the psf and pdb files
    lipidf : str
        The path to the folder containing the single-lipid PDB structures as well as the lipid DB file
    ff : list[str]
        AMBER forcefield files for lipids
    topo : list[str]
        AMBER topologies for lipids
    param : list[str]
        AMBER parameters for lipids
    build : bool
        Build system with teLeap. Disable if you want to build the system yourself.

    Returns
    -------
    mol : :class:`Molecule <moleculekit.molecule.Molecule`
        The resulting membrane including surrounding waters

    Examples
    --------
    >>> lipidratioupper = {'popc': 10, 'chl1': 1}
    >>> lipidratiolower = {'popc': 8, 'chl1': 2}
    >>> width = [50, 100]
    >>> res = buildMembrane(width, lipidratioupper, lipidratiolower)
    """
    from htmd.membranebuilder.ringpenetration import resolveRingPenetrations
    from htmd.builder.solvate import solvate
    from htmd.builder import amber
    from htmd.util import tempname
    from moleculekit.molecule import Molecule
    from htmd.home import home
    import os
    import pandas as pd

    if isinstance(equilibrate, bool):
        raise ValueError("equilibrate must be a float")

    if equilibrate > 0 or minimize > 0:
        build = True

    if lipidf is None:
        lipidf = os.path.join(home(shareDir=True), "membranebuilder", "lipids")
    lipiddb = pd.read_csv(os.path.join(lipidf, "lipiddb.csv"), index_col="Name")

    uqlip = np.unique(list(ratioupper.keys()) + list(ratiolower.keys()))
    files = _locateLipidFiles(lipidf, uqlip)

    area = np.prod(xysize)
    lipids = _createLipids(ratioupper, area, lipiddb, files, leaflet="upper")
    lipids += _createLipids(ratiolower, area, lipiddb, files, leaflet="lower")

    _setPositionsLJSim(xysize, [ll for ll in lipids if ll.xyz[2] > 0])
    _setPositionsLJSim(xysize, [ll for ll in lipids if ll.xyz[2] < 0])

    _findNeighbours(lipids, xysize)

    _loadMolecules(lipids, files)

    # from globalminimization import minimize
    # newpos, newrot = minimize(lipids, xysize + [100], stepxy=0.5, steprot=50, contactthresh=1)
    # for i in range(len(lipids)):
    #     lipids[i].xyz[:2] = newpos[i]
    #     lipids[i].rot = newrot[i]

    resolveRingPenetrations(lipids, xysize)
    memb = _createMembraneMolecule(lipids)

    minc = memb.get("coords", "name P").min(axis=0) - 5
    maxc = memb.get("coords", "name P").max(axis=0) + 5

    mm = [[0, 0, maxc[2] - 2], [xysize[0], xysize[1], maxc[2] + waterbuff]]
    smemb = solvate(memb, minmax=mm)
    mm = [[0, 0, minc[2] - waterbuff], [xysize[0], xysize[1], minc[2] + 2]]
    smemb = solvate(smemb, minmax=mm)

    smemb.moveBy([0, 0, -smemb.coords[:, 2, 0].min()])

    if outdir is None:
        outdir = tempname()
        logger.info(f"Outdir {outdir}")

    if build:
        res = amber.build(
            smemb,
            ionize=False,
            outdir=outdir,
            ff=ff,
            topo=topo,
            param=param,
        )
    else:
        os.makedirs(outdir, exist_ok=True)
        smemb.write(os.path.join(outdir, "structure.pdb"))
        return smemb

    if equilibrate > 0 or minimize > 0:
        from shutil import move
        import yaml
        from acemd import acemd

        equildir = os.path.join(outdir, "equil")
        os.makedirs(equildir, exist_ok=True)

        watcoo = res.get("coords", "water")
        celld = (watcoo.max(axis=0) - watcoo.min(axis=0)).squeeze()

        # TODO: I need to figure out the resnames+atom names after building
        # lipid_heads = list(set([(lipid.resname, lipid.headname) for lipid in lipids]))

        input_dict = {
            "structure": os.path.join(outdir, "structure.prmtop"),
            "coordinates": os.path.join(outdir, "structure.pdb"),
            "boxsize": celld.tolist(),
            "barostatconstratio": True,
            "minimize": minimize,
            "run": f"{equilibrate}ns",
            "velocities": 300,
            "thermostat": True,
            "barostat": True,
            # "extforces": [
            #     {
            #         "type": "positionalrestraint",
            #         "sel": " and ".join(
            #             [f"(resname {x[0]} and name {x[1]})" for x in lipid_heads]
            #         ),
            #         "setpoints": ["1@0"],
            #     }
            # ],
        }
        with open(os.path.join(equildir, "input.yaml"), "w") as f:
            yaml.dump(input_dict, f)
        acemd(equildir, platform=platform)

        if equilibrate > 0:
            res = Molecule(os.path.join(outdir, "structure.prmtop"))
            res.read(os.path.join(equildir, "output.coor"))
            res.read(os.path.join(equildir, "output.xsc"))
            res.wrap("lipids", guessBonds=False)
        else:
            res = Molecule(os.path.join(outdir, "structure.prmtop"))
            res.read(os.path.join(equildir, "minimized.coor"))

        move(
            os.path.join(outdir, "structure.pdb"),
            os.path.join(outdir, "starting_structure.pdb"),
        )
        res.write(os.path.join(outdir, "structure.pdb"))

    return res


def _findLeastAreaLipid(folder):
    """
    Use this to select the single least stretched conformation from a CHARMM-GUI lipid library
    """
    from glob import glob
    from scipy.spatial.distance import cdist

    ff = glob(folder + "/*/*.crd")
    maxdist = []
    for f in ff:
        m = Molecule(f)
        center = m.coords.mean(axis=0)
        dists = cdist(m.coords[:, :2, 0], center[:2].T)
        maxdist.append(dists.max())
    return ff[np.argmin(maxdist)], np.min(maxdist)

from htmd.builder.amber import _resolve_backend
try:
    _resolve_backend()
    tleap_installed = True
except Exception:
    tleap_installed = False


class _TestBuildMembrane(unittest.TestCase):
    def test_build_membrane(self):
        import tempfile
        import os

        with tempfile.TemporaryDirectory() as tmpdir:
            buildMembrane(
                [20, 20],
                ratioupper={"popc": 0.42, "pope": 0.4, "chl1": 0.18},
                ratiolower={"popc": 0.42, "pope": 0.4, "chl1": 0.18},
                equilibrate=0,
                minimize=0,
                outdir=tmpdir,
                platform="CPU",
            )
            self.assertTrue(os.path.exists(os.path.join(tmpdir, "structure.pdb")))
            self.assertFalse(
                os.path.exists(os.path.join(tmpdir, "starting_structure.pdb"))
            )

    @unittest.skipIf(
        not tleap_installed, "teLeap is not installed. Cannot test"
    )
    def test_build_membrane_minimize(self):
        import tempfile
        import os

        with tempfile.TemporaryDirectory() as tmpdir:
            buildMembrane(
                [20, 20],
                ratioupper={"popc": 0.42, "pope": 0.4, "chl1": 0.18},
                ratiolower={"popc": 0.42, "pope": 0.4, "chl1": 0.18},
                equilibrate=0,
                minimize=100,
                outdir=tmpdir,
                platform="CPU",
            )
            self.assertTrue(os.path.exists(os.path.join(tmpdir, "structure.pdb")))
            self.assertTrue(
                os.path.exists(os.path.join(tmpdir, "starting_structure.pdb"))
            )

    @unittest.skipIf(
        not tleap_installed, "teLeap is not installed. Cannot test"
    )
    def test_build_membrane_equil(self):
        import tempfile
        import os

        with tempfile.TemporaryDirectory() as tmpdir:
            buildMembrane(
                [20, 20],
                ratioupper={"popc": 0.42, "pope": 0.4, "chl1": 0.18},
                ratiolower={"popc": 0.42, "pope": 0.4, "chl1": 0.18},
                equilibrate=0.01,
                minimize=100,
                outdir=tmpdir,
                platform="CPU",
            )
            self.assertTrue(os.path.exists(os.path.join(tmpdir, "structure.pdb")))
            self.assertTrue(
                os.path.exists(os.path.join(tmpdir, "starting_structure.pdb"))
            )


if __name__ == "__main__":
    unittest.main()



# (c) 2015-2022 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import numpy as np
import logging

logger = logging.getLogger(__name__)


def least_square_fit_plane(coords):
    com = coords.mean(axis=0)
    numatm = coords.shape[0]

    axes = np.zeros((numatm, 3))
    for i in range(numatm):
        p1 = coords[i]
        if i == numatm - 1:
            p2 = coords[0]
        else:
            p2 = coords[i + 1]
        a = np.cross(p1, p2)
        axes += a
    u, d, v = np.linalg.svd(axes)
    return v[0], com


def getAllNeighbours(lipids, idx):
    neigh = list(lipids[idx].neighbours)
    for j, l in enumerate(lipids):
        if idx in l.neighbours:
            neigh.append(j)
    return neigh


def wrap(coor, com2, box):
    com1 = coor.mean(axis=0)
    for i in range(2):
        if (com1[i] - com2[i]) > (box[i] / 2):
            coor[:, i] -= box[i]
        if (com1[i] - com2[i]) < -(box[i] / 2):
            coor[:, i] += box[i]
    return coor


def moveLipidToPos(mol, lip):
    from moleculekit.util import rotationMatrix

    mol = mol.copy()
    headpos = mol.coords[mol.name == lip.headname].flatten()[np.newaxis, :]
    mol.moveBy(-headpos)
    mol.rotateBy(rotationMatrix([0, 0, 1], np.deg2rad(lip.rot)))
    mol.moveBy(lip.xyz)
    return np.squeeze(mol.coords[:, :, 0])


def _getRingProperties(ringcoords):
    axis, com = least_square_fit_plane(ringcoords)

    # project atoms to the least square fit plane
    for i, atom in enumerate(ringcoords):
        w = np.dot(axis, atom - com) * axis + com
        ringcoords[i] = com + (atom - w)

    maxd = np.max(np.sqrt(np.sum(np.square(ringcoords - com), axis=1)))
    return axis, com, maxd, ringcoords


def _checkAtomsPenetrating(coords2, bonds2, ringcoords, axis, ringcom, maxd, box):
    # Taken from https://github.com/sunhwan/lipid-pentest
    # find two bonded atoms that are at the opposite side of the plane
    coords2 = wrap(coords2, ringcom, box)

    d = np.sqrt(np.sum(np.square(coords2 - ringcom), axis=1))
    closeidx = np.where(d < 3)[0]
    if len(closeidx) == 0:
        return

    flag = False
    penetrating = None
    for i in closeidx:
        bonded = np.hstack((bonds2[bonds2[:, 0] == i, 1], bonds2[bonds2[:, 1] == i, 0]))
        for j in bonded:
            crd1 = coords2[i, :]
            crd2 = coords2[j, :]
            v1 = np.dot(crd1 - ringcom, axis)
            v2 = np.dot(crd2 - ringcom, axis)
            if v1 * v2 > 0:
                continue

            # point of intersection of the least square fit plane
            s = -np.dot(axis, crd1 - ringcom) / np.dot(axis, crd2 - crd1)
            p = crd1 + s * (crd2 - crd1)

            d = np.sqrt(np.sum(np.square(p - ringcom)))
            if d > maxd:
                continue

            d = 0
            for k in range(0, ringcoords.shape[0]):
                p1 = ringcoords[k] - p
                try:
                    p2 = ringcoords[k + 1] - p
                except Exception:
                    p2 = ringcoords[0] - p
                d += np.arccos(np.dot(p1, p2) / np.linalg.norm(p1) / np.linalg.norm(p2))

            wn = d / 2 / np.pi
            if 1.1 > wn > 0.9:
                penetrating = [i, j]
                flag = True
                break
        if flag:
            break
    return penetrating


def _detectRingPenetration(l1, lipids, box):
    lip1 = lipids[l1]
    neighbours = getAllNeighbours(lipids, l1)
    mol1 = lip1.mol
    coords1 = moveLipidToPos(mol1, lip1)
    bonds1 = mol1._getBonds()

    for l2, lip2 in zip(neighbours, lipids[neighbours]):
        if lip2.rings is None:
            continue
        mol2 = lip2.mol
        coords2 = moveLipidToPos(mol2, lip2)

        for r, ring in enumerate(lip2.rings):
            axis, ringcom, maxd, ringcoords = _getRingProperties(coords2[ring, :])

            pen = _checkAtomsPenetrating(
                coords1, bonds1, ringcoords, axis, ringcom, maxd, box
            )
            if pen is not None:
                logger.info(
                    "Lipid {} ring {} is being penetrated by lipid {} atoms {} {}".format(
                        l2, r, l1, pen[0], pen[1]
                    )
                )
                return True
    return False


def resolveRingPenetrations(lipids, box):
    lipids = np.array(lipids)
    while True:
        penetrators = []
        for l1 in range(len(lipids)):
            if _detectRingPenetration(l1, lipids, box):
                penetrators.append(l1)
        for p in penetrators:
            lipids[p].rot += 10
        if len(penetrators) == 0:
            logger.info("{} penetrating molecule(s) remaining".format(len(penetrators)))
            break
        logger.info("{} penetrating molecule(s) remaining".format(len(penetrators)))



# (c) 2015-2022 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import numpy as np
import logging
from sklearn.base import BaseEstimator, ClusterMixin, TransformerMixin

logger = logging.getLogger(__name__)


class RegCluster(BaseEstimator, ClusterMixin, TransformerMixin):
    """Class to perform regular clustering of a given data set

    RegCluster can be passed a radius or an approximate number of clusters. If a number of clusters is passed, KCenter
    clustering is used to estimate the necessary radius. RegCluster randomly chooses a point and assigns all points
    within the radius of this point to the same cluster. Then it proceeds with the nearest point, which is not yet
    assigned to a cluster and puts all unassigned points within the radius of this point in the next cluster and so on.

    Parameters
    ----------
    radius: float
        radius of clusters
    n_clusters: int
        desired number of clusters

    Examples
    --------
    >>> cluster = RegCluster(radius=5.1)
    >>> cluster.fit(data)

    Attributes
    ----------
    cluster_centers:  list
        list with the points, which are the centers of the clusters
    centerFrames : list
        list of indices of center points in data array
    labels_ : list
        list with number of cluster of each frame
    clusterSize_ : list
        list with number of frames in each cluster
    """

    def __init__(self, radius=None, n_clusters=None):
        if radius is None and n_clusters is None:
            raise RuntimeError("radius or n_clusters needs to be set")

        self.radius = radius
        self.n_clusters = n_clusters
        self.labels_ = []

    def fit(self, data):
        """performs clustering of data

        Parameters
        ----------
        data: np.ndarray
                array of data points to cluster
        merge: int
                minimal number of frames within each cluster. Smaller clusters are merged into next big one
        """
        # if n_clusters is given and no r, estimate n_clusters
        if self.radius is None:
            from htmd.clustering.kcenters import KCenter

            estClust = KCenter(n_clusters=self.n_clusters)
            estClust.fit(data)
            self.radius = estClust.distance.max()
            logger.info(f"Estimated radius = {self.radius}")

        from deeptime.clustering import RegularSpace

        self._reg = RegularSpace(dmin=self.radius)
        self.labels_ = self._reg.fit_fetch(data).transform(data).flatten()

    @property
    def cluster_centers_(self):
        return self._reg.model.cluster_centers

    @property
    def clusterSize(self):
        return np.bincount(self.labels_)


#
# class RegCluster:
#     """ Class to perform regular clustering of a given data set
#
#     RegCluster can be passed a radius or an approximate number of clusters. If a number of clusters is passed, KCenter
#     clustering is used to estimate the necessary radius. RegCluster randomly chooses a point and assigns all points
#     within the radius of this point to the same cluster. Then it proceeds with the nearest point, which is not yet
#     assigned to a cluster and puts all unassigned points within the radius of this point in the next cluster and so on.
#
#     Parameters
#     ----------
#     radius: float
#         radius of clusters
#     n_clusters: int
#         desired number of clusters
#
#     Examples
#     --------
#     >>> cluster = RegCluster(radius=5.1)
#     >>> cluster.fit(data)
#
#     Attributes
#     ----------
#     cluster_centers:  list
#         list with the points, which are the centers of the clusters
#     centerFrames : list
#         list of indices of center points in data array
#     labels_ : list
#         list with number of cluster of each frame
#     clusterSize_ : list
#         list with number of frames in each cluster
#     """
#     def __init__(self, radius=None, n_clusters=None):
#         if radius is None and n_clusters is None:
#             raise RuntimeError("radius or n_clusters needs to be set")
#
#         self.radius = radius
#         self.cluster_centers_ = []
#         self.centerFrames = []
#         self.labels_ = []
#         self.clusterSize = []
#         self.n_clusters = n_clusters
#
#     def fit(self, data):
#         """ performs clustering of data
#
#         Parameters
#         ----------
#         data: np.ndarray
#                 array of data points to cluster
#         merge: int
#                 minimal number of frames within each cluster. Smaller clusters are merged into next big one
#         """
#
#         if len(self.cluster_centers_) != 0:
#             logger.warning('Clustering already exists. Reclustering data!')
#
#         # if n_clusters is given and no r, estimate n_clusters
#         if self.radius is None:
#             estClust = KCenter(n_clusters=self.n_clusters)
#             estClust.fit(data)
#             self.radius = estClust.distance.max()
#             logger.info("Estimated radius = {}".format(self.radius))
#
#         self.cluster_centers_ = []
#         self.centerFrames = []
#         self.labels_ = np.ones(data.shape[0], dtype=int) * -1
#
#         # find initial center
#         unassigned = np.where(self.labels_ == -1)[0]
#         idxCenter = rd.choice(unassigned)
#         self.cluster_centers_.append(data[idxCenter, :])
#         self.centerFrames.append(idxCenter)
#
#         # find all points within radius from initial center
#         dist = self._dist(self.cluster_centers_, data)
#         self.labels_[dist <= self.radius] = 0
#
#         countCluster = 1
#
#         # Progress bar will not work here
#         while len(np.where(self.labels_ == -1)[0]) > 0:
#             if np.mod(countCluster, 100) == 0:
#                 logger.info('{} clusters created'.format(countCluster))
#             # find closest unassigned  point to existing clusters
#             unassigned = np.atleast_1d(np.where(self.labels_ == -1)[0])
#             newCenterIdx = unassigned[np.argmin(dist[unassigned])]
#             newCenter = data[newCenterIdx, :]
#
#             # make it a center
#             self.cluster_centers_.append(newCenter)
#             self.centerFrames.append(newCenterIdx)
#
#             # find unassigned points within radius from new cluster
#             newdist = self._dist(self.cluster_centers_[-1], data[unassigned, :])
#             self.labels_[unassigned[newdist <= self.radius]] = countCluster
#             countCluster += 1
#
#             # Update minimum distances
#             smallerdist = newdist < dist[unassigned]
#             dist[unassigned[smallerdist]] = newdist[smallerdist]
#
#         self.cluster_centers_ = np.array(self.cluster_centers_)
#         self.clusterSize = np.bincount(self.labels_)
#
#     @staticmethod
#     def _dist(centers, data):
#         dist = np.squeeze(cdist(np.atleast_2d(data), np.atleast_2d(centers)))
#         return np.atleast_1d(dist)

if __name__ == "__main__":
    X = np.random.random((100, 2))
    reg = RegCluster(radius=0.4)
    reg.fit(X)

    # from matplotlib import pylab as plt
    # plt.figure()
    # plt.scatter(X[:, 0], X[:, 1], c=reg.labels_)
    # centers = np.array(reg.cluster_centers_)
    # plt.scatter(centers[:, 0], centers[:, 1], s=100, c=range(centers.shape[0]))
    # plt.show()



# (c) 2015-2022 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import numpy as np
from copy import deepcopy
import uuid
import h5py
import pickle
import unittest
import logging

logger = logging.getLogger(__name__)


def _getsizes(x):
    if x is not None:
        if isinstance(x, np.ndarray):
            return x.shape[0]
        else:
            return len(x)


class Trajectory(object):
    """Class used for storing trajectory projections, their clustering and state assignments.

    Parameters
    ----------
    projection : numpy.ndarray
        The projected metrics
    reference : numpy.ndarray
        Reference indices to the simulations and frames that generated the metrics
    sim : :class:`Sim <htmd.simlist.Sim>` object
        A simulation object
    cluster : numpy.ndarray
        A list of cluster indexes for each frame in the simulation
    """

    def __init__(self, projection=None, reference=None, sim=None, cluster=None):
        self._projection = projection
        self._reference = reference
        self._cluster = cluster
        self.sim = sim
        self._checkframes((self.projection, self.reference, self.cluster))

    def toHDF5(self, h5group: h5py.Group):
        h5group.create_dataset("_projection", data=self._projection)
        h5group.create_dataset("_reference", data=self._reference)
        if self._cluster is not None:
            h5group.create_dataset("_cluster", data=self._cluster)
        simgroup = h5group.create_group("sim")
        self.sim.toHDF5(simgroup)

    @staticmethod
    def fromHDF5(h5group: h5py.Group):
        from htmd.simlist import Sim

        self = Trajectory()
        self._projection = np.array(h5group["_projection"])
        self._reference = np.array(h5group["_reference"])
        if "_cluster" in h5group:
            self._cluster = np.array(h5group["_cluster"])
        if "sim" in h5group:
            self.sim = Sim.fromHDF5(h5group["sim"])
        return self

    @property
    def projection(self):
        """The projected metrics of this simulation trajectory"""
        return self._projection

    @projection.setter
    def projection(self, value):
        self._checkframes((value, self.reference, self.cluster))
        self._projection = value

    @property
    def reference(self):
        """The reference indices to the simulations and frames that generated the projections"""
        return self._reference

    @reference.setter
    def reference(self, value):
        self._checkframes((self.projection, value, self.cluster))
        self._reference = value

    @property
    def cluster(self):
        """The cluster indexes for each frame in the simulation"""
        return self._cluster

    @cluster.setter
    def cluster(self, value):
        self._checkframes((self.projection, self.reference, value))
        self._cluster = value

    @property
    def numFrames(self):
        """The number of frames in this trajectory"""
        return self.projection.shape[0]

    @property
    def numDimensions(self):
        """The number of dimensions in the projection of this trajectory"""
        return self.projection.shape[1]

    def _numframes(self, args):
        return np.unique([x for x in list(map(_getsizes, args)) if x is not None])

    def _checkframes(self, args):
        if len(self._numframes(args)) > 1:
            raise RuntimeError(
                "projection, reference and cluster must have same lengths"
            )

    def dropFrames(self, frames):
        """Drop frames from the trajectory

        Parameters
        ----------
        frames : list
            A list of frame indexes to drop
        """
        if np.min(frames) < 0 or np.max(frames) >= self.numFrames:
            raise RuntimeError(
                "Frames to drop must be > 0 and < {}".format(self.numFrames)
            )
        if self._projection is not None:
            self._projection = np.delete(self._projection, frames, axis=0)
        if self._reference is not None:
            self._reference = np.delete(self._reference, frames, axis=0)
        if self._cluster is not None:
            self._cluster = np.delete(self._cluster, frames, axis=0)
        self._checkframes((self.projection, self.reference, self.cluster))

    def copy(self):
        """Produces a deep copy of the object"""
        return deepcopy(self)

    def __repr__(self):
        return (
            "<{}.{} object at {}>\n".format(
                self.__class__.__module__, self.__class__.__name__, hex(id(self))
            )
            + self.__str__()
        )

    def __str__(self):
        return "sim: {}\nprojection: {}\nreference: {}\ncluster: {}\n".format(
            "simid = {}".format(self.sim.simid) if self.sim is not None else None,
            (
                "np.array(shape={})".format(np.shape(self.projection))
                if self.projection is not None
                else None
            ),
            (
                "np.array(shape={})".format(np.shape(self.reference))
                if self.reference is not None
                else None
            ),
            (
                "np.array(shape={})".format(np.shape(self.cluster))
                if self.cluster is not None
                else None
            ),
        )


class MetricData(object):
    """Class used for storing projected trajectories, their clustering and state assignments. Objects of this class
    are constructed by the `project` methods of the other projection classes. Only construct this class if you want to
    load saved data.

    Parameters
    ----------
    dat : numpy.ndarray
        The projected metrics
    ref : numpy.ndarray
        Reference indices to the simulations and frames that generated the metrics
    description : pandas.DataFrame
        A description of the metrics
    simlist : numpy.ndarray of :class:`Sim <htmd.simlist.Sim>` objects
        A simulation list generated by the :func:`simlist <htmd.simlist.simlist>` function
    fstep : float
        Size of simulation step in ns
    parent : :class:`MetricData` object
        The MetricData object that was used to generate this object
    file : str
        Path to a saved MetricData object
    trajectories : list of :class:`Trajectory` objects
        A list of Trajectory objects
    cluster : numpy.ndarray
        A list of cluster indexes for each frame in the data

    Attributes
    ----------
    trajectories : list of :class:`Trajectory` objects
        A list of Trajectory objects
    fstep : float
        Size of simulation step in ns
    description : pandas.DataFrame
        A description of the metrics
    K : int
        Number of clusters
    N : numpy.ndarray
        Number of frames in each cluster
    Centers : numpy.ndarray
        Cluster centers (if available)
    """

    def __init__(
        self,
        dat=None,
        ref=None,
        description=None,
        simlist=None,
        fstep=0,
        parent=None,
        file=None,
        trajectories=None,
        cluster=None,
    ):
        if trajectories is None:
            self._loadTrajectories(dat, ref, simlist, cluster)
        else:
            self.trajectories = trajectories

        self.fstep = fstep
        self.description = description
        self.parent = parent
        self.K = None
        self.N = None
        self.Centers = None

        if file is not None:
            self.load(file)

        self._dataid = str(uuid.uuid4())
        self._clusterid = None
        return

    def _loadTrajectories(
        self, projection=None, reference=None, simlist=None, cluster=None
    ):
        size = np.unique(
            [
                x
                for x in list(map(_getsizes, (projection, reference, simlist, cluster)))
                if x is not None
            ]
        )
        if len(size) == 0:
            size = 0
        elif len(size) > 1:
            raise RuntimeError("dat, ref and simlist must have same lengths")
        if projection is None:
            projection = np.empty(size, dtype=object)
        if reference is None:
            reference = np.empty(size, dtype=object)
        if simlist is None:
            simlist = np.empty(size, dtype=object)
        if cluster is None:
            cluster = np.empty(size, dtype=object)
        self.trajectories = [
            Trajectory(d, r, s, c)
            for d, r, s, c in zip(projection, reference, simlist, cluster)
        ]

    @property
    def dat(self):
        return [t.projection for t in self.trajectories]

    @property
    def ref(self):
        return [t.reference for t in self.trajectories]

    @property
    def St(self):
        return [t.cluster for t in self.trajectories]

    @property
    def simlist(self):
        return [x.sim for x in self.trajectories]

    @property
    def map(self):
        return self.description

    @property
    def trajLengths(self):
        """Get the lengths of all trajectories

        Returns
        -------
        lens : list
            The lengths of all trajectories in the object

        Examples
        --------
        >>> data.trajLengths
        """
        return np.array([x.numFrames for x in self.trajectories])

    @property
    def numFrames(self):
        """Get the total number of frames in all trajectories

        Returns
        -------
        nframes : int
            Total number of frames in all trajectories

        Examples
        --------
        >>> data.numFrames
        """
        return sum(self.trajLengths)

    @property
    def numTrajectories(self):
        """The number of trajectories

        Examples
        --------
        >>> data.numTrajectories
        """
        return len(self.trajectories)

    @property
    def numDimensions(self):
        """The number of dimensions

        Examples
        --------
        >>> data.numDimensions
        """
        return self.trajectories[0].numDimensions

    @property
    def aggregateTime(self):
        """The total aggregate simulation time

        Examples
        --------
        >>> data.aggTime
        """
        if self.fstep > 0:
            return self.numFrames * self.fstep

    def cluster(self, clusterobj, mergesmall=None, batchsize=False):
        """Cluster the metrics

        Parameters
        ----------
        clusterobj : :class:`ClusterMixin <sklearn.cluster.ClusterMixin>` object
            The object of a clustering class from sklearn or with the same interface
        mergesmall : int
            Clusters containing less than `mergesmall` conformations will be joined into their closest well-populated
            neighbour.
        batchsize : int
            Batch sizes bigger than 0 will enable batching.

        Examples
        --------
        >>> from sklearn.cluster import MiniBatchKMeans
        >>> data = MetricDistance.project(sims, 'protein and name CA', 'resname MOL')
        >>> data.cluster(MiniBatchKMeans(n_clusters=1000), mergesmall=5)
        """
        # cluster_obj = coor.cluster_kmeans(self.dat, k=20, stride=1)
        if batchsize > 0:
            lengths = self.trajLengths
            currsum = 0
            starts = [0]
            for i, l in enumerate(lengths):
                currsum += l
                if currsum > batchsize:
                    starts.append(i + 1)
                    currsum = 0
            starts.append(self.numTrajectories)
            for i in range(len(starts) - 1):
                clusterobj.partial_fit(
                    np.concatenate(self.dat[starts[i] : starts[i + 1]])
                )
            labels = []
            for i in range(self.numTrajectories):
                labels.append(clusterobj.predict(self.dat[i].astype(np.float32)))
            # This is retarded
            labels = np.concatenate(labels)
        else:
            datconcat = np.concatenate(self.dat)
            if np.ndim(datconcat) == 1:
                datconcat = np.transpose(np.atleast_2d(datconcat))
            import warnings  # Following 3 lines are BS because sklearn refuse to make releases more often than 1 per year...

            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                clusterobj.fit(datconcat)
            labels = clusterobj.labels_

        uqclu = np.unique(labels)
        self.Centers = clusterobj.cluster_centers_[uqclu, :]
        self.K = len(uqclu)
        # ---- Fixing missing clusters crap...
        map = np.zeros(np.max(labels) + 1, dtype=int) * -1
        map[uqclu] = range(self.K)
        labels = map[labels]
        # ------------------------------------
        for i, s in enumerate(self.deconcatenate(labels)):
            self.trajectories[i].cluster = s
        self.N = np.bincount(labels)

        if mergesmall is not None:
            oldK = self.K
            self.K, St, self.Centers, self.N, xxx = _mergeSmallClusters(
                mergesmall, datconcat, labels, self.Centers, self.N
            )
            for i, s in enumerate(self.deconcatenate(St)):
                self.trajectories[i].cluster = s
            logger.info(
                "Mergesmall removed {} clusters. Original ncluster {}, new ncluster {}.".format(
                    oldK - self.K, oldK, self.K
                )
            )

        self._dataid = str(uuid.uuid4())
        self._clusterid = self._dataid

    def combine(self, otherdata):
        """Combines two different metrics into one by concatenating them.

        Parameters
        ----------
        otherdata : :class:`MetricData` object
            Concatenates the metrics of otherdata to the current objects metrics

        Examples
        --------
        >>> dataRMSD = MetricRmsd.project(sims)
        >>> dataDist = MetricSelfDistance.project(sims, 'protein and name CA')
        >>> dataRMSD.combine(dataDist)
        """
        import pandas as pd

        if not np.array_equal(self.trajLengths, otherdata.trajLengths):
            raise NameError(
                "Trying to combine MetricData objects with different number/lengths of trajectories. Check the trajLengths property."
            )
        for i in range(self.numTrajectories):
            if self.simlist[i].simid != otherdata.simlist[i].simid:
                raise NameError(
                    "Simulation ids do not match. Cannot combine. Please generate both data from the same simlist"
                )
        for t1, t2 in zip(self.trajectories, otherdata.trajectories):
            t1.projection = np.concatenate((t1.projection, t2.projection), axis=1)
        self.description = pd.concat(
            [self.description, otherdata.description], ignore_index=True
        )
        self._dataid = str(uuid.uuid4())

    def dropDimensions(self, drop=None, keep=None):
        """Drop some dimensions of the data given their indexes

        Parameters
        ----------
        drop : list
            A list of integer indexes of the dimensions to drop
        keep : list
            A list of integer indexes of the dimensions to keep

        Examples
        --------
        >>> data.dropDimensions([1, 24, 3])
        >>> data.dropDimensions(keep=[2, 10])
        """
        if drop is not None and not isinstance(drop, np.ndarray):
            drop = np.array(drop)
        if keep is not None and not isinstance(keep, np.ndarray):
            keep = np.array(keep)
        if drop is not None and keep is not None:
            raise AttributeError(
                "drop and keep arguments for dropDimensions are mutually exclusive. Pass only one."
            )
        if keep is not None:
            keepidx = keep
            dropidx = np.arange(self.numDimensions)
            dropidx = np.setdiff1d(dropidx, keepidx)
        else:
            dropidx = drop
            keepidx = np.arange(self.numDimensions)
            keepidx = np.setdiff1d(keepidx, dropidx)

        for t in self.trajectories:
            t.projection = t.projection[:, keepidx]
        self.description = self.description.drop(self.description.index[dropidx])
        self.description = self.description.reset_index(drop=True)

    def dropTraj(
        self, limits=None, multiple=None, partial=None, idx=None, keepsims=None
    ):
        """Drops trajectories based on their lengths

        By default, drops all trajectories which are not of statistical mode (most common) length.

        Parameters
        ----------
        limits : list, optional
            Lower and upper limits of trajectory lengths we want to keep. e.g. [100, 500]
        multiple : list, optional
            Drops trajectories whose length is not a multiple of lengths in the list. e.g. [50, 80]
        partial : bool
            Not implemented yet
        idx : list, optional
            A list of trajectory indexes to drop
        keepsims : list of :class:`Sim <htmd.simlist.Sim>` objects
            A list of sims which we want to keep

        Examples
        --------
        >>> data = MetricSelfDistance.project(sims, 'protein and name CA')
        >>> data.dropTraj()
        >>> data.dropTraj(multiple=[100])
        """
        trajLengths = self.trajLengths
        orgNum = self.numTrajectories

        if limits is not None:
            drop = (trajLengths < limits[0]) | (trajLengths > limits[1])
        elif multiple is not None:
            if partial is not None:
                raise NameError("TODO")
                pass
            idx = range(orgNum)
            for i in range(len(multiple)):
                idxNew = np.where(np.mod(trajLengths, multiple[i]) != 0)
                idx = np.intersect1d(idxNew, idx)
            drop = np.zeros(orgNum, dtype=bool)
            drop[idx] = True
        elif idx is not None:
            drop = np.zeros(orgNum, dtype=bool)
            drop[idx] = True
        elif keepsims is not None:
            # Fast check to see if simlists are identical
            if len(keepsims) == self.numTrajectories:
                allsame = True
                for i, t in enumerate(self.trajectories):
                    if keepsims[i] != t.sim:
                        allsame = False
                        break
                if allsame:
                    return

            # Slow check where all sims are checked against each other
            drop = np.ones(orgNum, dtype=bool)
            for i, s in enumerate(self.simlist):
                for k in keepsims:
                    if s == k:
                        drop[i] = False
                        break
        else:
            from scipy import stats

            drop = trajLengths != np.array(stats.mode(trajLengths, keepdims=False).mode)

        keep = np.invert(drop)
        dropIdx = np.where(drop)[0]

        self.trajectories = [self.trajectories[x] for x in np.where(keep)[0]]
        self._dataid = str(uuid.uuid4())
        if self.parent:
            self.parent.trajectories = [
                self.parent.trajectories[x] for x in np.where(keep)[0]
            ]
            self.parent._dataid = str(uuid.uuid4())

        logger.info(
            "Dropped "
            + str(np.sum(drop))
            + " trajectories from "
            + str(orgNum)
            + " resulting in "
            + str(self.numTrajectories)
        )
        return dropIdx

    def dropFrames(self, idx, frames):
        self.trajectories[idx].dropFrames(frames)
        self._dataid = str(uuid.uuid4())
        if self.parent:
            self.parent.trajectories[idx].dropFrames(frames)
            self.parent._dataid = str(uuid.uuid4())

    def sampleClusters(
        self, clusters=None, frames=20, replacement=False, allframes=False
    ):
        """Samples frames from a set of clusters

        Parameters
        ----------
        clusters : Union[None, list]
            A list of cluster indexes from which we want to sample
        frames : Union[None, int, list]
            An integer with the number of frames we want to sample from each state or a list of same length as
            `clusters` which contains the number of frames we want from each of the clusters.
            If None is given it will return all frames.
        replacement : bool
            If we want to sample with or without replacement.
        allframes : bool
            Deprecated. Use frames=None instead.

        Returns
        -------
        absframes : numpy.ndarray
            An array which contains for each state an array containing absolute trajectory frames
        relframes : numpy.ndarray
            An array which contains for each state a 2D array containing the trajectory ID and frame number for each of
            the sampled frames

        Examples
        --------
        >>> data.sampleClusters(range(5), [10, 3, 2, 50, 1])  # Sample from first 5 clusters, 10, 3, etc frames respectively
        """
        if clusters is None:
            clusters = range(self.K)
        if isinstance(clusters, int):
            clusters = [
                clusters,
            ]
        if allframes:
            logger.warning("allframes option is deprecated. Please use frames=None")
            frames = None
        if frames is None or isinstance(frames, int):
            frames = np.repeat(frames, len(clusters))

        stConcat = np.concatenate(self.St)
        absFrames = []
        relFrames = []
        for i in range(len(clusters)):
            if frames[i] == 0 and not allframes:
                continue
            st = clusters[i]
            absFrames.append(_sampleCluster(st, stConcat, frames[i], replacement))
            if len(absFrames[-1]) == 0:
                raise NameError(
                    "No frames could be sampled from cluster {}. Cluster is empty.".format(
                        st
                    )
                )

            relFrames.append(self.abs2rel(absFrames[-1]))
        return absFrames, relFrames

    def bootstrap(self, ratio, replacement=False):
        """Randomly sample a set of trajectories

        Parameters
        ----------
        ratio : float
            What ratio of trajectories to keep. e.g. 0.8
        replacement : bool
            If we should sample with replacement

        Returns
        -------
        bootdata : :class:`MetricData` object
            A new :class:`MetricData` object containing only the sampled trajectories

        Examples
        --------
        >>> data = MetricSelfDistance.project(sims, 'protein and name CA')
        >>> databoot = data.bootstrap(0.8)
        """
        numtraj = self.numTrajectories
        numtokeep = int(np.floor(numtraj * ratio))
        if replacement:
            rndtraj = np.random.randint(numtraj, size=numtokeep)
        else:
            rndtraj = np.random.permutation(numtraj)[0:numtokeep]
        rndtraj = sorted(
            rndtraj
        )  # Important to keep the sorting! i.e. for data.dropTraj(keepsims=sims)

        pp = None
        if self.parent is not None:
            pp = self.parent.copy()
            pp.trajectories = [self.parent.trajectories[x] for x in rndtraj]
            pp._dataid = str(uuid.uuid4())
        bootdata = MetricData(
            trajectories=[self.trajectories[x].copy() for x in rndtraj],
            description=self.description,
            parent=pp,
            fstep=self.fstep,
        )
        return bootdata

    def plotTrajSizes(self):
        """Plot the lengths of all trajectories in a sorted bar plot

        Examples
        --------
        >>> data = MetricSelfDistance.project(sims, 'protein and name CA')
        >>> data.plotTrajSizes()
        """
        trajLengths = self.trajLengths * self.fstep
        import matplotlib.pyplot as plt

        plt.ion()
        _ = plt.hist(trajLengths)
        # plt.bar(range(len(trajLengths)), np.sort(trajLengths), color='b', edgecolor='b')
        plt.ylabel("Num trajectories")
        plt.xlabel("Length of trajectories (in ns)")
        plt.show()
        return

    def splitCols(self):
        raise NameError("Not implemented yet")

    def deconcatenate(self, array):
        indeces = np.cumsum(self.trajLengths)
        if np.ndim(array) == 1:
            return np.split(array, indeces[:-1])
        else:
            return np.vsplit(array, indeces[:-1])

    def abs2rel(self, absFrames):
        """Convert absolute frame indexes into trajectory index-frame pairs

        Useful when doing calculations on a concatenated data array of all trajectories. When you find a frame of
        interest you can `deconcatenate` the frame index to the corresponding trajectory index-frame pair.

        Parameters
        ----------
        absFrames : list of int
            A list of absolute index frames

        Returns
        -------
        pairs : np.ndarray
            A array where each row is a trajectory index-frame pair

        Examples
        --------
        >>> relidx = data.abs2rel(536)
        """
        if not hasattr(absFrames, "__len__"):
            absFrames = [absFrames]
        endFrames = np.append(0, np.cumsum(self.trajLengths))

        relframe = np.zeros((len(absFrames), 2), dtype=int)
        for i in range(len(absFrames)):
            trajIdx = np.where(absFrames[i] < endFrames)[0][0] - 1
            trajFr = absFrames[i] - endFrames[trajIdx]
            relframe[i, :] = [trajIdx, trajFr]
        return relframe

    def rel2sim(self, relFrames, simlist=None):
        """Converts trajectory index-frame pairs into Sim-frame pairs

        Parameters
        ----------
        relFrames : 2D np.ndarray
            An array containing in each row trajectory index and frame pairs
        simlist : numpy.ndarray of :class:`Sim <htmd.simlist.Sim>` objects
            Optionally pass a different (but matching, i.e. filtered) simlist for creating the Frames.

        Returns
        -------
        frames : np.ndarray
            An array of :class:`Frame <htmd.simlist.Frame>` objects containing the simulation object, the trajectory
            piece ID and the frame index.

        Examples
        --------
        >>> simframes = data.rel2sim([100, 56])  # 100th simulation frame 56
        """
        from htmd.simlist import Frame

        if simlist is None:
            simlist = self.simlist
        else:
            if len(simlist) != len(self.simlist):
                raise AttributeError(
                    "Provided simlist has different number of trajectories than the one used by this object."
                )

        relFrames = np.array(relFrames)
        if relFrames.ndim == 1:
            relFrames = relFrames[np.newaxis, :]

        frames = []
        for trajID, trajFrame in relFrames:
            ref = self.trajectories[trajID].reference
            frames.append(Frame(simlist[trajID], ref[trajFrame, 0], ref[trajFrame, 1]))
        return np.array(frames)

    def abs2sim(self, absFrames):
        """Converts absolute frame indexes into Sim-frame pairs

        Parameters
        ----------
        absFrames : list of int
            A list of absolute index frames

        Returns
        -------
        frames : np.ndarray
            An array of :class:`Frame <htmd.simlist.Frame>` objects containing the simulation object, the trajectory
            piece ID and the frame index.

        Examples
        --------
        >>> simframes = data.abs2sim(563)  # 563rd frame to simulation/frame pairs
        """
        return self.rel2sim(self.abs2rel(absFrames))

    def copy(self):
        """Produces a deep copy of the object

        Returns
        -------
        data : :class:`MetricData` object
            A copy of the current object

        Examples
        --------
        >>> data = MetricSelfDistance.project(sims, 'protein and name CA')
        >>> data2 = data.copy()
        """
        return deepcopy(self)

    def toHDF5(self, filename=None, h5group=None):
        import h5py

        if h5group is None and filename is None:
            raise RuntimeError("Either h5group or filename has to be set")

        h5f = None
        if h5group is None:
            h5f = h5py.File(filename, "w")
            h5group = h5f.create_group("MetricData")

        def _set_attrs(grp, keys, attr=True):
            for key in keys:
                if getattr(self, key):
                    if attr:
                        grp.attrs[key] = getattr(self, key)
                    else:
                        grp.create_dataset(key, data=getattr(self, key))

        _set_attrs(h5group, ["K", "_dataid", "_clusterid", "fstep"])
        h5group.attrs["description"] = self.description.to_json()
        _set_attrs(h5group, ["N", "Centers"], False)

        trajgrp = h5group.create_group("trajectories")
        for i in range(len(self.trajectories)):
            trajg = trajgrp.create_group(str(i))
            self.trajectories[i].toHDF5(trajg)

        if self.parent is not None:
            parentgroup = h5group.create_group("parent")
            self.parent.toHDF5(h5group=parentgroup)

        if h5f is not None:
            h5f.close()

    @staticmethod
    def fromHDF5(filename=None, h5group=None):
        import pandas as pd
        from natsort import natsorted

        if h5group is None and filename is None:
            raise RuntimeError("Either h5group or filename has to be set")

        data = MetricData()

        h5f = None
        if h5group is None:
            h5f = h5py.File(filename, "r")
            h5group = h5f["MetricData"]

        def _get_attrs(grp, keys, attr=True):
            for key in keys:
                if attr and key in grp.attrs:
                    setattr(data, key, grp.attrs[key])
                elif key in grp:
                    setattr(data, key, np.array(grp[key]))

        _get_attrs(h5group, ["K", "_dataid", "_clusterid", "fstep"])
        data.description = pd.read_json(h5group.attrs["description"])
        _get_attrs(h5group, ["N", "Centers"], False)

        data.trajectories = []
        if "trajectories" in h5group:
            for key in natsorted(h5group["trajectories"]):
                data.trajectories.append(
                    Trajectory.fromHDF5(h5group["trajectories"][key])
                )

        if "parent" in h5group:
            data.parent = MetricData.fromHDF5(h5group=h5group["parent"])

        if h5f is not None:
            h5f.close()

        return data

    def save(self, filename):
        """Save a :class:`MetricData` object to disk

        Parameters
        ----------
        filename : str
            Path of the file in which to save the object

        Examples
        --------
        >>> data = MetricSelfDistance.project(sims, 'protein and name CA')
        >>> data.save('./data.dat')
        """
        # np.save(filename, [self.__dict__[k] for k in self.__dict__])
        parentpointer = self.parent
        if self.parent is not None:
            self.parent = self.parent.__dict__

        f = open(filename, "wb")
        pickle.dump(self.__dict__, f)
        f.close()

        if self.parent is not None:
            self.parent = parentpointer

    def load(self, filename):
        """Load a :class:`MetricData` object from disk

        Parameters
        ----------
        filename : str
            Path to the saved MetricData object

        Examples
        --------
        >>> data = MetricData()
        >>> data.load('./data.dat')
        """
        import sys

        try:
            import pandas.indexes
        except ImportError:
            import pandas.core.indexes

            sys.modules["pandas.indexes"] = (
                pandas.core.indexes
            )  # Hacky fix for new pandas version

        # Patch for old HTMD versions
        if type(filename).__name__ == "MetricData":
            filename = filename.__dict__

        if isinstance(filename, str):
            f = open(filename, "rb")
            vardict = pickle.load(f)
            f.close()
        elif isinstance(filename, dict):
            vardict = filename

        for k in self.__dict__:
            if k == "description" and "map" in vardict:  # Patch for loading old data
                self.description = vardict["map"]
            elif k == "trajectories" and "dat" in vardict:  # Patch for loading old data
                self._loadTrajectories(
                    vardict["dat"], vardict["ref"], vardict["simlist"], vardict["St"]
                )
            elif k != "parent":
                try:
                    self.__dict__[k] = vardict[k]
                except Exception:
                    logger.warning(
                        "Could not find class property {} in file {}".format(
                            k, filename
                        )
                    )

        if "parent" in vardict and vardict["parent"] is not None:
            self.parent = MetricData()
            self.parent.load(vardict["parent"])

    def _defaultLags(self, minlag=None, maxlag=None, numlags=None, units="frames"):
        from htmd.units import convert as unitconvert

        if maxlag is None:
            from scipy import stats

            maxlag = (
                stats.mode(self.trajLengths, keepdims=False).mode - 1
            )  # -1 to avoid warnings in timescales calc
        else:
            maxlag = unitconvert(units, "frames", maxlag, fstep=self.fstep)

        if minlag is None:
            if maxlag > 20:
                minlag = 10
            else:
                minlag = 2
        else:
            minlag = unitconvert(units, "frames", minlag, fstep=self.fstep)

        return np.append(1, np.round(np.linspace(minlag, maxlag, numlags))).astype(int)

    def _getFEShistogramCounts(
        self,
        dimx,
        dimy,
        nbins=80,
        pad=0.5,
        micro_ofcluster=None,
        stationary_distribution=None,
        St=None,
    ):
        from tqdm import tqdm

        if St is None:
            St = self.St

        clusters = np.hstack(St)
        data_dim = []
        for traj in self.trajectories:
            data_dim.append(traj.projection[:, [dimx, dimy]])
        data_dim = np.vstack(data_dim)

        xmin, ymin = data_dim.min(axis=0)
        xmax, ymax = data_dim.max(axis=0)
        hist_range = np.array([[xmin - pad, xmax + pad], [ymin - pad, ymax + pad]])

        if micro_ofcluster is None:
            counts, xbins, ybins = np.histogram2d(
                data_dim[:, 0], data_dim[:, 1], bins=nbins, range=hist_range
            )
            return counts.T, xbins, ybins

        counts = np.zeros((nbins, nbins))
        for m in tqdm(range(len(stationary_distribution))):
            frames = micro_ofcluster[clusters] == m
            if frames.sum():
                state_mask, xbins, ybins = np.histogram2d(
                    data_dim[frames, 0],
                    data_dim[frames, 1],
                    bins=nbins,
                    range=hist_range,
                )
                state_mask = state_mask.T > 0
                counts += state_mask * stationary_distribution[m]
        return counts, xbins, ybins

    def _contourPlot(
        self,
        values,
        xbins,
        ybins,
        levels=7,
        nonzero=None,
        cmap="viridis",
        title=None,
        xlabel=None,
        ylabel=None,
    ):
        import matplotlib.pylab as plt

        if nonzero is None:
            nonzero = np.ones_like(values).astype(bool)

        def getcmap(cmap, bounds):
            import matplotlib.colors as clr

            cmap = plt.get_cmap(cmap)
            cmaplist = [cmap(i) for i in range(cmap.N)]
            cmap = clr.LinearSegmentedColormap.from_list(
                "Custom cmap", cmaplist, cmap.N
            )
            norm = clr.BoundaryNorm(bounds, cmap.N)
            cmap.set_under(color="white")
            cmap.set_over(color="white")
            return cmap, norm

        xcenters = (xbins[:-1] + xbins[1:]) / 2
        ycenters = (ybins[:-1] + ybins[1:]) / 2
        meshx, meshy = np.meshgrid(xcenters, ycenters)

        levels = np.linspace(values[nonzero].min(), values[nonzero].max(), levels)
        cmap, norm = getcmap(cmap, levels)
        f = plt.figure()
        plt.contour(
            meshx, meshy, values, levels=levels, colors="black", vmin=0, vmax=levels[-1]
        )
        cf = plt.contourf(
            meshx,
            meshy,
            values,
            levels=levels,
            cmap=cmap,
            vmin=0,
            vmax=levels[-1],
            norm=norm,
        )

        ax = f.gca()
        _ = ax.axis("equal")
        if title is not None:
            ax.set_title(title)
        if xlabel is not None:
            ax.set_xlabel(xlabel)
        if ylabel is not None:
            ax.set_ylabel(ylabel)
        return f, ax, cf

    def _setColorbar(self, f, mappable, label=None, scientific=True):
        import matplotlib.ticker as ticker

        def fmt(x, pos):
            a, b = "{:.2e}".format(x).split("e")
            b = int(b)
            return r"${} \times 10^{{{}}}$".format(a, b)

        if scientific:
            f.colorbar(mappable, format=ticker.FuncFormatter(fmt), label=label)
        else:
            f.colorbar(mappable, label=label)

    def _plotCounts(
        self,
        dimX,
        dimY,
        resolution=100,
        logplot=False,
        levels=7,
        cmap="viridis",
        title=None,
        xlabel=None,
        ylabel=None,
    ):
        counts, xbins, ybins = self._getFEShistogramCounts(dimX, dimY, nbins=resolution)
        nonzero = counts != 0
        if logplot:
            counts[nonzero] = np.log(counts[nonzero])
            nonzero = counts != 0
        f, ax, cf = self._contourPlot(
            counts,
            xbins,
            ybins,
            levels=levels,
            nonzero=nonzero,
            cmap=cmap,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
        )

        self._setColorbar(f, cf, "Counts")
        return f, ax, cf

    def plotCounts(
        self,
        dimX,
        dimY,
        resolution=100,
        logplot=False,
        plot=True,
        save=None,
        levels=7,
        cmap="viridis",
    ):
        """Plots a histogram of counts on any two given dimensions.

        Parameters
        ----------
        dimX : int
            Index of projected dimension to use for the X axis.
        dimY : int
            Index of projected dimension to use for the Y axis.
        resolution : int
            Resolution of bincount grid.
        logplot : bool
            Set True to plot the logarithm of counts.
        plot : bool
            If the method should display the plot
        save : str
            Path of the file in which to save the figure
        """
        from matplotlib import pylab as plt

        if self.description is not None:
            xlabel = self.description.description[dimX]
        else:
            xlabel = "Dimension {}".format(dimX)
        if self.description is not None:
            ylabel = self.description.description[dimY]
        else:
            ylabel = "Dimension {}".format(dimY)
        title = "Counts histogram"
        if logplot:
            title = "Logarithmic counts histogram"

        self._plotCounts(
            dimX,
            dimY,
            resolution=resolution,
            logplot=logplot,
            levels=levels,
            cmap=cmap,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
        )

        if save is not None:
            plt.savefig(save, dpi=300, bbox_inches="tight", pad_inches=0.2)
        if plot:
            plt.show()

    def plotClusters(
        self,
        dimX,
        dimY,
        resolution=100,
        s=4,
        c=None,
        cmap="Greys",
        logplot=False,
        plot=True,
        save=None,
        data=None,
        levels=7,
    ):
        """Plot a scatter-plot of the locations of the clusters on top of the count histogram.

        Parameters
        ----------
        dimX : int
            Index of projected dimension to use for the X axis.
        dimY : int
            Index of projected dimension to use for the Y axis.
        resolution : int
            Resolution of bincount grid.
        s : float
            Marker size for clusters.
        c : list
            Colors or indexes for each cluster.
        cmap : matplotlib.colors.Colormap
            Matplotlib colormap for the scatter plot.
        logplot : bool
            Set True to plot the logarithm of counts.
        plot : bool
            If the method should display the plot
        save : str
            Path of the file in which to save the figure
        data : :class:`MetricData` object
            Optionally you can pass a different MetricData object than the one used for clustering. For example
            if the user wants to cluster on distances but wants to plot the centers on top of RMSD values. The new
            MetricData object needs to have the same simlist as this object.
        """
        if self.Centers is None:
            raise RuntimeError("Data has not been clustered yet. Cannot plot clusters.")
        from matplotlib import pylab as plt

        if data is None:
            data = self
            centers = self.Centers
        else:
            from htmd.model import getStateStatistic

            if self.numFrames != data.numFrames or ~np.all(
                [s1 == s2 for s1, s2 in zip(self.simlist, data.simlist)]
            ):
                raise RuntimeError(
                    "The data argument you provided uses a different simlist than this object."
                )
            centers = np.vstack(
                getStateStatistic(self, data, range(self.K), statetype="cluster")
            )

        if data.description is not None:
            xlabel = data.description.description[dimX]
        else:
            xlabel = "Dimension {}".format(dimX)

        if data.description is not None:
            ylabel = data.description.description[dimY]
        else:
            ylabel = "Dimension {}".format(dimY)

        title = "Clusters plotted onto counts histogram"
        if logplot:
            title = "Clusters plotted onto logarithmic counts histogram"
        f, ax, cf = self._plotCounts(
            dimX,
            dimY,
            resolution=resolution,
            logplot=logplot,
            levels=levels,
            cmap=cmap,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
        )

        y = ax.scatter(
            centers[:, dimX],
            centers[:, dimY],
            s=s,
            c=c if c is not None else "r",
            cmap=cmap,
            linewidths=0,
            marker="o",
        )
        if c is not None:
            self._setColorbar(f, y, "Cluster groups")

        if save is not None:
            plt.savefig(save, dpi=300, bbox_inches="tight", pad_inches=0.2)
        if plot:
            plt.show()

    def __repr__(self):
        return (
            "<{}.{} object at {}>\n".format(
                self.__class__.__module__, self.__class__.__name__, hex(id(self))
            )
            + self.__str__()
        )

    def __str__(self):
        def formatstr(name, field):
            if isinstance(field, np.ndarray) or isinstance(field, list):
                rep = "{} shape: {}".format(name, np.shape(field))
            else:
                rep = "{}: {}".format(name, field)
            return rep

        rep = "MetricData object with {} trajectories".format(self.numTrajectories)
        if self.fstep > 0:
            rep += " of {:.1f}ns aggregate simulation time".format(self.aggregateTime)
        for j in sorted(self.__dict__.keys()):
            if j[0] == "_":
                continue
            if j == "parent":
                rep += "\nparent: {} at {}".format(
                    type(self.parent), hex(id(self.parent))
                )
            elif j == "description":
                rep += "\ndescription: {} at {}".format(
                    type(self.description), hex(id(self.description))
                )
            else:
                rep += "\n"
                rep += formatstr(j, self.__dict__[j])

        return rep

    def append(self, other):
        self.trajectories += other.trajectories
        for i, t in enumerate(self.trajectories):
            t.sim.simid = i

        if self.parent and other.parent:
            self.parent.trajectories += other.parent.trajectories
            for i, t in enumerate(self.parent.trajectories):
                t.sim.simid = i

        self._dataid = str(uuid.uuid4())
        self._resetClustering()
        return self

    def _resetClustering(self):
        for t in self.trajectories:
            t.cluster = None
        self.K = None
        self.N = None
        self.Centers = None

    def sampleRegion(
        self, point=None, radius=None, limits=None, nsamples=20, singlemol=False
    ):
        """Samples conformations from a region in the projected space.

        Parameters
        ----------
        point : list or np.ndarray
            A point in the projected space. Undefined dimensions should have None value.
        radius : float
            The radius in around the point in which to sample conformations.
        limits : np.ndarray
            A (2, ndim) dimensional array containing the min (1st row) and max (2nd row) limits for each dimension.
            None values will be interpreted as no limit in that dimension, or min/max value.
        nsamples : int
            The number of conformations to sample.
        singlemol : bool
            If True it will return all samples within a single Molecule instead of a list of Molecules.

        Returns
        -------
        absFrames : list
            A list of the absolute frame indexes sampled
        relFrames : list of tuples
            A list of (trajNum, frameNum) tuples sampled
        mols : Molecule or list of Molecules
            The conformations stored in a Molecule or a list of Molecules

        Examples
        --------
        >>> # Working with 4 dimensional data for example
        >>> abs, rel, mols = data.sampleRegion(point=(0.5, 3, None, None), radius=0.1)  # Point undefined in dim 3, 4
        >>> minlims = [-1, None, None, 4]  # No min limit for 2, 3 dim
        >>> maxlims = [2,     3, None, 7]  # No max limit for 3 dim
        >>> abs, rel, mols = data.sampleRegion(limits=np.array([minlims, maxlims]))
        """
        from scipy.spatial.distance import cdist

        datconcat = np.concatenate(self.dat)
        numdim = datconcat.shape[1]
        if point is not None:
            if radius is None:
                raise RuntimeError("You must define a radius with a point.")
            point = np.array(point)
            if len(point) != numdim:
                raise RuntimeError(
                    "Argument `point` should be same dimensionality as your data ({} dimensions)".format(
                        numdim
                    )
                )
            keepdim = np.array([p is not None for p in point])
            dists = cdist(datconcat[:, keepdim], [point[keepdim]])
            confs = np.where(dists < radius)[0]
        elif limits is not None:
            if limits.shape != (2, numdim):
                raise RuntimeError(
                    "Argument `limits` should be of shape (2, {})".format(numdim)
                )
            mask = np.ones(datconcat.shape[0], dtype=bool)
            for i in range(numdim):
                if limits[0, i] is not None:
                    mask &= datconcat[:, i] > limits[0, i]
                if limits[1, i] is not None:
                    mask &= datconcat[:, i] < limits[1, i]
            confs = np.where(mask)[0]

        if len(confs) > nsamples:
            confs = np.random.choice(confs, nsamples, replace=False)
        sims = self.abs2sim(confs)

        from moleculekit.molecule import Molecule

        if singlemol:
            mol = Molecule(sims[0])
            for i in range(1, len(sims)):
                m = Molecule(sims[i])
                mol.appendFrames(m)
        else:
            mol = []
            for s in sims:
                mol.append(Molecule(s))
        return confs, self.abs2rel(confs), mol


def _sampleCluster(cluster, stConcat, numFrames, replacement):
    frames = np.where(stConcat == cluster)[0]
    return _randomSample(frames, numFrames, replacement)


def _randomSample(frames, numFr, replacement):
    if numFr == 0:
        return []
    if numFr is None or (numFr >= len(frames) and not replacement):
        rnd = list(range(len(frames)))
    else:
        rnd = np.random.randint(len(frames), size=numFr)
    return frames[rnd]


def _mergeSmallClusters(mergesmall, data, stconcat, centers, N, metric=None):
    if data.dtype == "bool":
        metric = "hamming"
    else:
        metric = "euclidean"
    badclusters = N < mergesmall
    goodclusters = np.invert(badclusters)
    N[badclusters] = 0
    badcluidx = np.where(badclusters)[0]
    goodcluidx = np.where(goodclusters)[0]

    if len(badcluidx) == 0:
        return len(N), stconcat, centers, N, badclusters

    # Keep only good centers
    centers = centers[goodclusters, :]

    # Creating a mapping for new cluster numbers as we will have to remove some
    newidx = np.zeros(len(N), dtype=int)
    newidx[goodcluidx] = range(len(goodcluidx))

    # Find all frames which belong to bad clusters
    frames = _ismember(stconcat, badcluidx)
    badframeidx = np.where(frames >= 0)[0]

    # Calculate distance of all frames belonging to bad clusters to the good cluster centers
    from scipy.spatial import distance

    dists = distance.cdist(
        np.atleast_2d(data[badframeidx, :]), np.atleast_2d(centers), metric
    )
    minidx = np.argmin(
        dists, axis=1
    )  # Find closest center. Indexes are relative to goodidx
    newclu = goodcluidx[minidx]  # Back to absolute cluster indexes

    # Reassign bad frames to good clusters
    stconcat[badframeidx] = newclu  # Assign them to new clusters
    stconcat = newidx[stconcat]  # Convert all cluster indexes to the new indexes

    N = np.bincount(stconcat)
    K = len(N)
    return K, stconcat, centers, N, badclusters


def _ismember(a, b):
    bind = {}
    for i, elt in enumerate(list(set(b))):
        bind[elt] = i
    return np.array(
        [bind.get(itm, -1) for itm in a]
    )  # None can be replaced by any other "not in b" value


def _generate_toy_data(trans_prob, n_traj=1000, n_frames=1000, seed=None, cluster=True):
    from htmd.metricdata import MetricData
    from pandas import DataFrame

    assert np.all((np.sum(trans_prob, axis=1)) - 1 < 1e-15)

    # Create real transition probability matrix
    n_states = trans_prob.shape[0]

    # Generate fake trajectories from it
    dat = []
    ref = []
    if seed is not None:
        np.random.seed(seed)

    for i in range(n_traj):  # ntraj
        trajref = np.zeros((n_frames, 2), dtype=np.int32)
        trajref[:, 0] = i
        trajref[:, 1] = np.arange(n_frames)
        trajdat = np.zeros((n_frames, n_states), dtype=np.int32)
        curr_state = np.random.randint(0, n_states)
        trajdat[0, curr_state] = 1
        for j in range(1, n_frames):  # nsteps
            curr_state = np.random.choice(n_states, p=trans_prob[curr_state])
            trajdat[j, curr_state] = 1
        dat.append(trajdat)
        ref.append(trajref)

    # Create fake test data
    data = MetricData(dat=dat, ref=ref)
    data._dataid = "fake"
    data.fstep = 1

    if cluster:
        for traj in data.trajectories:
            traj.cluster = np.where(traj.projection)[1].copy()
        data._clusterid = "fake"
        data.K = n_states
        data.N = np.bincount(np.concatenate(data.St))
        data.Centers = np.zeros((n_states, n_states), dtype=np.int32)

    types = ["coordinate"] * n_states
    indexes = np.arange(n_states)
    description = ["coordinate"] * n_states
    data.description = DataFrame(
        {"type": types, "atomIndexes": indexes, "description": description}
    )

    return data


class _TestMetricData(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        from htmd.simlist import simlist, simfilter
        from glob import glob
        from htmd.projections.metric import Metric
        from moleculekit.projections.metricdistance import MetricDistance
        from moleculekit.projections.metricdihedral import MetricDihedral
        from moleculekit.util import tempname
        from htmd.home import home
        from os.path import join

        sims = simlist(
            glob(join(home(dataDir="adaptive"), "data", "*", "")),
            glob(join(home(dataDir="adaptive"), "input", "*")),
        )
        fsims = simfilter(sims, tempname(), "not water")

        metr = Metric(fsims)
        metr.set(
            MetricDistance(
                "protein and resid 10 and name CA",
                "resname BEN and noh",
                periodic="selections",
                metric="contacts",
                groupsel1="residue",
                threshold=4,
            )
        )
        self.data1 = metr.project()

        metr.set(MetricDihedral())
        self.data2 = metr.project()

    def test_combine(self):
        # Testing combining of metrics
        data1 = self.data1.copy()
        data1.combine(self.data2)

        # Testing dimensions
        assert np.array_equal(
            data1.description.shape, (897, 3)
        ), "combine not working correct"
        assert np.array_equal(
            data1.trajectories[0].projection.shape, (6, 897)
        ), "combine not working correct"
        assert np.array_equal(
            np.where(data1.description.type == "contact")[0],
            [0, 1, 2, 3, 4, 5, 6, 7, 8],
        ), "combine not working correct"

    def test_dropping(self):
        # Testing dimension dropping / keeping
        data1 = self.data1.copy()
        assert np.array_equal(data1.description.shape, (9, 3))
        data1.dropDimensions(range(9))
        assert np.array_equal(
            data1.description.shape, (0, 3)
        ), "dropDimensions not working correct"
        assert np.array_equal(
            data1.trajectories[0].projection.shape, (6, 0)
        ), "dropDimensions not working correct"
        assert (
            len(np.where(data1.description.type == "contact")[0]) == 0
        ), "dropDimensions not working correct"

        data2 = self.data2.copy()
        assert np.array_equal(data2.description.shape, (888, 3))
        data2.dropDimensions(keep=range(9))
        assert np.array_equal(
            data2.description.shape, (9, 3)
        ), "dropDimensions not working correct"
        assert np.array_equal(
            data2.trajectories[0].projection.shape, (6, 9)
        ), "dropDimensions not working correct"
        assert (
            len(np.where(data2.description.type == "dihedral")[0]) == 9
        ), "dropDimensions not working correct"

    def test_saving_loading(self):
        from moleculekit.util import tempname

        def checkCorrectness(newdata):
            assert newdata.numTrajectories == 2, "Failed to load trajectories"
            assert newdata.description.shape == (9, 3), "Failed to load pandas data"
            assert newdata.trajectories[0].projection.shape == (6, 9), "Wrong data size"

        savefile = tempname(suffix=".dat")
        self.data1.save(savefile)

        newdata = MetricData(file=savefile)
        checkCorrectness(newdata)

        newdata = MetricData()
        newdata.load(savefile)
        checkCorrectness(newdata)

        # Saving with a parent
        data1 = self.data1.copy()
        data1.parent = self.data2.copy()
        data1.save(savefile)
        newdata = MetricData(file=savefile)
        checkCorrectness(newdata)


if __name__ == "__main__":
    import unittest

    unittest.main(verbosity=2)



# (c) 2015-2022 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import numpy as np
from moleculekit.molecule import Molecule
from htmd.metricdata import MetricData
from scipy import stats
from moleculekit.projections.projection import Projection
from joblib import Parallel, delayed
from htmd.parallelprogress import ParallelExecutor
import unittest
import logging

logger = logging.getLogger(__name__)


class Metric:
    """Class for calculating projections of a simlist.

    Parameters
    ----------
    simulations : list
        A list of simulations produced by :func:`simlist <htmd.simlist.simlist>`
    skip : int
        Frame skipping. Setting i.e. to 3 will keep only every third frame of each simulation.
    metricdata : :class:`MetricData <htmd.metricdata.MetricData>` object
        If a MetricData object is passed in the constructor, Metric will try to update it by only adding simulations
        which don't exist in it yet.

    Examples
    --------
    >>> metr = Metric(sims)  # doctest: +SKIP
    >>> metr.set(MetricSelfDistance('protein and name CA', metric='contacts'))  # doctest: +SKIP
    >>> data = metr.project()  # doctest: +SKIP
    >>>
    >>> # Or define your own function which accepts as first argument a Molecule object. Further arguments are passed as
    >>> # function/argument tuples
    >>> def foo(mol, ref):
    >>>     from moleculekit.util import molRMSD
    >>>     mol.wrap('protein')
    >>>     mol.align('protein and name CA', refmol=ref)
    >>>     return molRMSD(mol, ref, mol.atomselect('protein and name CA'), ref.atomselect('protein and name CA'))
    >>>
    >>> metr = Metric(sims)
    >>> metr.set( (foo, (ref,)) )
    >>> data2 = metr.project()

    .. rubric:: Methods
    .. autoautosummary:: htmd.projections.metric.Metric
        :methods:
    .. rubric:: Attributes
    .. autoautosummary:: htmd.projections.metric.Metric
        :attributes:
    """

    def __init__(self, simulations, skip=1, metricdata=None):
        self.simulations = simulations
        self.skip = skip
        self.projectionlist = []
        self.metricdata = metricdata

    def set(self, projection):
        """Sets the projection to be applied to the simulations.

        Parameters
        ----------
        projection : function or :class:`Projection <moleculekit.projections.projection.Projection>` object or list of objects
            A function or projection or a list of projections/functions which to use on the simulations
        """
        self.projectionlist = projection
        if not (
            isinstance(self.projectionlist, list)
            or isinstance(self.projectionlist, tuple)
        ) and (
            isinstance(self.projectionlist, Projection)
            or hasattr(self.projectionlist, "__call__")
        ):
            self.projectionlist = [
                self.projectionlist,
            ]
        elif isinstance(self.projectionlist, list) or isinstance(
            self.projectionlist, tuple
        ):
            if hasattr(self.projectionlist[0], "__call__"):
                self.projectionlist = [
                    self.projectionlist,
                ]
            pass
        else:
            raise AttributeError(
                "Metric.set only accepts Projection objects, functions, function/argument tuples or "
                "lists and tuples thereof."
            )

    def getMapping(self, mol):
        """Returns the description of each projected dimension.

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object which will be used to calculate the descriptions of the projected dimensions.

        Returns
        -------
        map : :class:`DataFrame <pandas.core.frame.DataFrame>` object
            A DataFrame containing the descriptions of each dimension
        """
        import pandas as pd

        if mol is None:
            return
        pandamap = pd.DataFrame(columns=("type", "atomIndexes", "description"))
        for proj in self.projectionlist:
            if isinstance(proj, Projection):
                pandamap = pd.concat(
                    [pandamap, proj.getMapping(mol)], ignore_index=True
                )
        return pandamap

    def project(self, njobs=None):
        """
        Applies all projections stored in Metric on all simulations.

        Parameters
        ----------
        njobs : int
            Number of parallel jobs to spawn for projection of trajectories. Take care that this can use large amounts
            of memory as multiple trajectories are loaded at once.  If None it will use the default from htmd.config.

        Returns
        -------
        data : MetricData object
               Returns a MetricData object containing the projected data.
        """
        if len(self.projectionlist) == 0:
            raise RuntimeError(
                "You need to provide projections using the Metric.set method."
            )

        # Projecting single Molecules
        if isinstance(self.simulations, Molecule):
            data = []
            mol = self.simulations
            for proj in self.projectionlist:
                data.append(_project(proj, mol))
            return data

        numSim = len(self.simulations)

        # Find out if there is a unique molfile. If there is, initialize a single Molecule to speed up calculations
        uqMol = None
        (single, molfile) = _singleMolfile(self.simulations)
        if single:
            uqMol = Molecule(molfile)
            for proj in self.projectionlist:
                if isinstance(proj, Projection):
                    proj._setCache(uqMol)
        else:
            logger.warning(
                "Cannot calculate description of dimensions due to different topology files for each trajectory."
            )
        mapping = self.getMapping(uqMol)

        logger.debug("Metric: Starting projection of trajectories.")
        from htmd.config import _config

        aprun = ParallelExecutor(
            n_jobs=njobs if njobs is not None else _config["njobs"]
        )
        results = aprun(total=numSim, desc="Projecting trajectories")(
            delayed(_processSim)(
                self.simulations[i], self.projectionlist, uqMol, self.skip
            )
            for i in range(numSim)
        )

        metrics = []
        ref = []
        deletesims = np.zeros(numSim, dtype=bool)
        fstep = np.zeros(numSim)
        for i in range(len(results)):
            metrics.append(results[i][0])
            ref.append(results[i][1])
            fstep[i] = results[i][2]
            deletesims[i] = results[i][3]

        logger.debug("Finished projecting the trajectories.")

        # Removing empty trajectories
        metrics, ref, updlist, fstep = self._removeEmpty(
            metrics, ref, deletesims, fstep
        )

        # Constructing a MetricData object
        data = MetricData(dat=metrics, ref=ref, description=mapping, simlist=updlist)

        uqfsteps = np.unique(fstep)
        if np.all(np.isnan(uqfsteps)):
            logger.warning(
                "No framestep could be read from the trajectories. Please manually set the MetricData.fstep"
                " property, otherwise calculations in Model and Kinetics classes can fail."
            )
        else:
            data.fstep = float(stats.mode(fstep, keepdims=False).mode)
            if len(uqfsteps) != 1:
                logger.warning(
                    "Multiple framesteps [{}] ns were read from the simulations. "
                    "Taking the statistical mode: {}ns. "
                    "If it looks wrong, you can modify it by manually "
                    "setting the MetricData.fstep property.".format(
                        ", ".join(map(str, uqfsteps)), data.fstep
                    )
                )
            else:
                if data.fstep == 0:
                    logger.warning(
                        "A framestep of 0 was read from the trajectories. Please manually set the MetricData.fstep"
                        " property, otherwise calculations in Model and Kinetics classes can fail."
                    )
                else:
                    logger.info(
                        "Frame step {}ns was read from the trajectories. If it looks wrong, redefine it by manually "
                        "setting the MetricData.fstep property.".format(data.fstep)
                    )

        return data

    def _projectSingle(self, index):
        data, ref, fstep, _ = _processSim(
            self.simulations[index], self.projectionlist, None, self.skip
        )
        return data, ref, fstep

    def _removeEmpty(self, metrics, ref, deletesims, fstep):
        emptyM = np.array([True if x is None else False for x in metrics], dtype=bool)
        emptyR = np.array([True if x is None else False for x in ref], dtype=bool)
        assert np.all(deletesims == emptyM) and np.all(emptyR == emptyM)

        idx = np.where(emptyM)[0]

        metrics = [metrics[i] for i in range(len(metrics)) if i not in idx]
        ref = [ref[i] for i in range(len(ref)) if i not in idx]
        updlist = [
            self.simulations[i] for i in range(len(self.simulations)) if i not in idx
        ]
        fstep = [fstep[i] for i in range(len(fstep)) if i not in idx]

        if len(metrics) == 0:
            raise NameError(
                "No trajectories were projected. Check if the simlist is empty or for projection errors."
            )

        return metrics, ref, updlist, fstep


def _project(proj, target):
    if isinstance(proj, Projection):
        res = proj.project(target)
    elif hasattr(proj, "__call__"):  # If it's a function
        res = proj(target)
    elif isinstance(proj, tuple) and hasattr(
        proj[0], "__call__"
    ):  # If it's a function with extra arguments
        res = proj[0](target, *proj[1])
    else:
        raise RuntimeError("Invalid projection type {}".format(type(proj)))

    if isinstance(res, list) or isinstance(res, tuple):
        res = np.array(res)

    _checkProjectionDims(res, target, "projection")
    return res


def _checkProjectionDims(result, mol, name):
    if (result.ndim == 1 and len(result) != mol.numFrames) or (
        result.ndim == 2 and result.shape[0] != mol.numFrames
    ):
        raise RuntimeError(
            "The {} produced an incorrect result. It produced a {} shaped array for {} frames. "
            "The {} should return a numpy array of (nframes, ndim) shape "
            "where nframes the number of frames in the Molecule it accepts as an "
            "argument.".format(name, result.shape, mol.numFrames, name)
        )


def _highfreqFilter(mol, steps):
    newframes = int(mol.coords.shape[2] / steps) * steps
    mol.coords = mol.coords[:, :, :newframes]
    mol.box = mol.box[:, :newframes]
    mol.box = mol.box[:, ::steps]
    mol.fstep = mol.fstep * steps
    coo = np.reshape(
        mol.coords,
        [
            mol.coords.shape[0],
            mol.coords.shape[1],
            int(mol.coords.shape[2] / steps),
            steps,
        ],
    )
    X = np.mean(coo, axis=3)
    #   print("mol: ", mol.coords.shape, " X: ",X.shape)
    mol.coords = X


def _processSim(sim, projectionlist, uqmol, skip):
    pieces = sim.trajectory
    try:
        if uqmol is not None:
            mol = uqmol.copy()
        else:
            mol = Molecule(sim.molfile)
        logger.debug(pieces[0])

        mol.read(pieces, skip=skip)

        data = []
        for proj in projectionlist:
            result = _project(proj, mol)
            if result.ndim == 1:
                result = np.atleast_2d(result).T
            if result.size == 0:
                logger.warning(
                    f"No data was produced by projection {proj.__class__} of simulation id: {sim.simid}"
                )
            data.append(result)

        data = np.hstack(data)
        if data.dtype == np.float64:
            data = data.astype(np.float32)

    except Exception as e:
        logger.warning(f'Error while projecting simulation id: {sim.simid}. "{e}"')
        return None, None, None, True

    return data, _calcRef(pieces, mol.fileloc), mol.fstep, False


def _calcRef(pieces, fileloc):
    locs = np.array([x[0] for x in fileloc])
    frames = list([x[1] for x in fileloc])
    ref = np.zeros((len(frames), 2), dtype="u4")
    ref[:, 1] = frames
    for i, p in enumerate(pieces):
        ref[locs == p, 0] = i
    return ref


def _singleMolfile(sims):
    from moleculekit.molecule import mol_equal
    from htmd.util import ensurelist

    if isinstance(sims, Molecule):
        return False, []
    elif isinstance(sims, (list, tuple, np.ndarray)):
        molfiles = []
        for s in sims:
            molfiles.append(tuple(ensurelist(s.molfile)))

        uqmolfiles = list(set(molfiles))

        if len(uqmolfiles) == 0:
            raise RuntimeError("No molfiles found in simlist")
        elif len(uqmolfiles) == 1:
            return True, uqmolfiles[0]
        elif (
            len(uqmolfiles) > 1
        ):  # If more than one molfile load them and see if they are different Molecules
            ref = Molecule(uqmolfiles[0], _logger=False)
            for i in range(1, len(uqmolfiles)):
                mol = Molecule(uqmolfiles[i], _logger=False)
                if not mol_equal(ref, mol, exceptFields=["coords"]):
                    return False, []
            return True, uqmolfiles[0]
    return False, []


def _projectionGenerator(metric, njobs):
    for i in range(0, len(metric.simulations), njobs):
        simrange = range(i, np.min((i + njobs, len(metric.simulations))))
        results = Parallel(n_jobs=njobs, verbose=0)(
            delayed(_projector)(metric, i) for i in simrange
        )
        yield results


def _projector(metric, i):
    return metric._projectSingle(i)


class _TestMetric(unittest.TestCase):
    def test_set(self):
        from moleculekit.molecule import Molecule
        from moleculekit.projections.metricrmsd import MetricRmsd

        # Testing the set method of Metric
        sims = []
        ref = Molecule("3PTB")
        metr = Metric(sims)
        metr.set(MetricRmsd(ref, "protein and name CA"))
        assert len(metr.projectionlist) == 1

        metr.set(
            [
                MetricRmsd(ref, "protein and name CA"),
                MetricRmsd(ref, "protein and name CA"),
            ]
        )
        assert len(metr.projectionlist) == 2

    def test_function_projections(self):
        def foo(mol, ref):
            from moleculekit.util import molRMSD

            mol.wrap("protein")
            mol.align("protein and name CA", refmol=ref)
            return molRMSD(
                mol,
                ref,
                mol.atomselect("protein and name CA"),
                ref.atomselect("protein and name CA"),
            )

        ref = Molecule("3PTB")
        sims = []
        metr = Metric(sims)
        metr.set((foo, (ref,)))
        assert len(metr.projectionlist) == 1


if __name__ == "__main__":
    unittest.main(verbosity=2)



# (c) 2015-2022 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#

from joblib import Parallel, delayed  # Import delayed as well for other modules

"""
A wrapper for joblib.Parallel to allow custom progress bars.
"""

from tqdm import tqdm


def ParallelExecutor(**joblib_args):
    def aprun(**tq_args):
        tqdm_f = lambda x, args: tqdm(x, **args)
        return lambda x: Parallel(**joblib_args)(tqdm_f(x, tq_args))

    return aprun



# (c) 2015-2022 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import os


def compareVersions():
    from htmd import __version__ as currver
    from natsort import natsorted
    import os
    import time
    from os.path import expanduser

    __home = expanduser("~")
    __htmdconf = os.path.join(__home, ".htmd")
    if not os.path.exists(__htmdconf):
        try:
            os.makedirs(__htmdconf)
        except Exception:
            print(
                f"Unable to create {__htmdconf} folder. Will not check for new HTMD versions."
            )
            return
    __file = os.path.join(__htmdconf, ".latestversion")

    # Check if one day has passed since last version check. If yes, get new version and write to file
    if (
        not os.path.isfile(__file)
        or time.time() > os.path.getmtime(__file) + 86400
        or os.stat(__file).st_size == 0
    ):
        _writeLatestVersionFile(__file)

    try:
        with open(__file, "r") as f:
            latest = f.read().strip()
    except Exception:
        print(
            f"Unable to open {__file} file for reading. Will not check for new HTMD versions."
        )
        return

    if natsorted((latest, currver))[1] != currver:
        print(
            f"New HTMD version ({latest}) is available. You are currently on ({currver}).\n"
            "We recommend you create a new conda environment with the latest HTMD version.\n"
            f"Run: `conda create -n htmd{latest} htmd={latest} -c acellera -c conda-forge`\n"
        )
    else:
        print(f"You are on the latest HTMD version ({currver}).")

    print("")


def _writeLatestVersionFile(fname):
    import urllib.request
    import json

    version = None
    try:
        url = "https://api.anaconda.org/package/acellera/htmd"
        with urllib.request.urlopen(url) as r:
            version = json.loads(r.read())["latest_version"]
    except Exception:
        return

    try:
        with open(fname, "w") as f:
            f.write(version)
    except Exception:
        print(
            f"Unable to open {fname} file for writing. Will not check for new HTMD versions."
        )
        return

    os.utime(fname, None)
